# Copyright (c) 2017 Ultimaker B.V.
# This example is released under the terms of the AGPLv3 or higher.
#import trimesh

from . import trimesh
from . import Jobs

from UM.Logger import Logger

from typing import cast, List, Optional

from UM.Scene.Iterator.DepthFirstIterator import DepthFirstIterator
from UM.View.SelectionPass import SelectionPass


from UM.Scene.SceneNode import SceneNode
from UM.Job import Job
from UM.Event import Event, MouseEvent, KeyEvent
from UM.Message import Message

from UM.Math.Plane import Plane
from UM.Math.Vector import Vector
from UM.Math.Quaternion import Quaternion
from UM.Math.Float import Float

from UM.Math.Color import Color
from cura.CuraApplication import CuraApplication
from cura.Scene.CuraSceneNode import CuraSceneNode
from cura.PickingPass import PickingPass

from UM.Operations.GroupedOperation import GroupedOperation
from UM.Operations.TranslateOperation import TranslateOperation
from UM.Operations.AddSceneNodeOperation import AddSceneNodeOperation
from UM.Operations.RemoveSceneNodeOperation import RemoveSceneNodeOperation
from cura.Operations.SetParentOperation import SetParentOperation

from cura.Scene.SliceableObjectDecorator import SliceableObjectDecorator
from cura.Scene.BuildPlateDecorator import BuildPlateDecorator

import numpy
import os.path #To find the QML files.
from PyQt5.QtCore import pyqtProperty, pyqtSignal, Qt, QUrl, QObject, QVariant #To define a shortcut key and to find the QML files, and to expose information to QML.
from PyQt5.QtQml import QQmlComponent, QQmlContext #To create a dialogue window.
from UM.Qt.ListModel import ListModel

from UM.Application import Application #To register the information dialogue.
from UM.PluginRegistry import PluginRegistry #To find the QML files in the plug-in folder.
from UM.Scene.Selection import Selection #To get the current selection and some information about it.
from UM.Tool import Tool #The PluginObject we're going to extend.
from UM.Scene.ToolHandle import ToolHandle


from UM.Operations.RotateOperation import RotateOperation
from UM.Operations.GroupedOperation import GroupedOperation
from UM.Operations.SetTransformOperation import SetTransformOperation
from UM.Operations.LayFlatOperation import LayFlatOperation

from UM.Version import Version

from UM.View.GL.OpenGL import OpenGL

from . import Operations
from . import PrettyNode

from UM.Mesh.MeshData import MeshData, calculateNormalsFromIndexedVertices

try:
    from . import AutoparterToolHandle
except (ImportError, SystemError):
    import AutoparterToolHandle  # type: ignore  # This fixes the tests not being able to import.

import math
import time

from UM.i18n import i18nCatalog
i18n_catalog = i18nCatalog("uranium")


from UM.Mesh.MeshBuilder import MeshBuilder
from enum import Enum

DIMENSION_TOLERANCE = 0.0001

class Mode(Enum):
    MOVE = 1
    SCALE_LEFT_CORNER = 2
    SCALE_RIGHT_CORNER = 3

class CamView(Enum):
    FRONTAL = 1
    TOP = 2
    BOTTOM = 3
    LEFT = 4
    RIGHT = 5
    CUSTOM = 6

class Autoparter(Tool):

    def __init__(self):
        super().__init__()

        self._handle = AutoparterToolHandle.AutoparterToolHandle()
        self._handle.setVisible(False)

        self._innerRectHandle = AutoparterToolHandle.InnerRectToolHandle()
        self._innerRectHandle.setVisible(False)

        self._leftCornerHandle = AutoparterToolHandle.LeftCornerToolHandle()
        self._rightCornerHandle = AutoparterToolHandle.RightCornerToolHandle()
        self._leftCornerHandle.setVisible(False)
        self._rightCornerHandle.setVisible(False)



        self._shortcut_key = Qt.Key_X

        self._activeView = "Welcome"
        self._activeSubView = 0

        #self._activeView = "CreateConnectors"
        #self._activeSubView = 2

        self.info_window = None

        self.setExposedProperties("Tolerance", "OneHalfSelected", "CameraView", "IsEditingCubicConnector", "CanAddCubicConnector", "CanAddShapedConnector", "CanCreateConnectors", "ActiveView", "AidActivated", "IsSlicing", "ShowError", "ShowDepthError", "ErrorMessage", "ActiveSubView", "ConnectorsList", "SelectedConnectorIndex", "ConnectorGender", "DepthA", "DepthB", "CanFinallyAddCubicConnector")

        self._showError = False
        self._errorMsg = ""

        self._oneHalfNode = None
        self._otherHalfNode = None

        ##
        # PlaneSelection
        ##



        self._planeStyle = 0
        # 0: plane is displayed red,
        # 1: plane is displayed black
        # 2: plane is displayed red / "white"
        self._sliceOperation = None
        self._selectedObject = None
        self._sectionalPlaneX = 0
        self._sectionalPlaneMesh = None
        self._sectionalPlaneAidMesh = None
        self._sectionalPlaneCuttingMesh = None
        self._sectionalPlaneToolHandle = None
        self._isHandleEnabled = False
        self._isSlicing = False
        self._intentionallyRemoved = False

        ##
        # CreateConnectors
        ##
        self._oneHalfCutSurfaceData = None
        self._otherHalfCutSurfaceData = None

        self._cameraView = CamView.FRONTAL.value

        self._notFrontalViewActive = False
        self._disabledNodes = []
        self._temporarilyRemovedNodes = []
        self._hiddenNodes = []
        self._unselectedNodes = []
        self._oneHalfSelected = True
        self._viewPosX = 0
        self._viewPosY = 0
        self._oneHalfNodeYOffset = 0
        self._otherHalfNodeYOffset = 0
        self._currentHandleYOffset = 0

        self._checksCollidingWithMesh = False
        self._showDepthError = False

        self._cutsurfaceNode = None
        self._cutsurfacePosA = None
        self._cutsurfacePosB = None

        self._cutSurfaceCenterPosition = None


        self._connectorsList = ListModel()

        self._connectorsList.addRoleName(Qt.UserRole + 1, "connectorType")
        self._connectorsList.addRoleName(Qt.UserRole + 2, "number")
        self._connectorsList.setItems([])
        self._selectedConnectorIndex = -1

        self._cubicConnectors = []
        self._areCubicConnectors3D = False

        self._isEditingCubicConnector = False

        self._innerHandleBorder = 0.2
        self._connectorGender = 0
        self._rectHandleWithinPoly = False
        self._rectHandleCollidesWithOtherCubicConnectors = True
        self._moved = False
        self._isRectInPolyJob = None
        self._isCollidingWithOtherCCJob = None
        self._isCollidingWithMeshJob = None
        self._depthA = 0.5
        self._depthB = 0.5
        self._tolerance = 0.1
        self._currentDepth = 0.5
        #
        self._mode = Mode.MOVE
        self._distance = None


    ################################
    ##
    # Events
    ##
    ################################

    def event(self, event) -> bool:
        #super().event(event) #The super event updates the selection if the user clicks on some object.


        if not self._selection_pass:

            self._selection_pass = cast(SelectionPass, Application.getInstance().getRenderer().getRenderPass("selection"))

            if not self._selection_pass:
                return False

        if event.type == Event.ToolActivateEvent and Selection.hasSelection():
            if(self._activeView == "PlaneSelection"):
                self._enableCutsurfaceHandle(0)


        if event.type == Event.ToolDeactivateEvent:

            if(self._activeView == "PlaneSelection"):
                self._enableCutsurfaceHandle(1)
            if(self._activeView == "CreateConnectors" and self._activeSubView == 1):
                #self._cancelPressed()
                Logger.log("i", "nth")

        if self._activeView == "CreateConnectors" and (self._activeSubView == 1 or self._activeSubView == 2):

            if event.type is event.KeyPressEvent:
                if event.key == KeyEvent.UpKey:
                    self._moveView(0,10)
                    return True
                if event.key == KeyEvent.DownKey:
                    self._moveView(0,-10)
                    return True
                if event.key == KeyEvent.RightKey:
                    self._moveView(10,0)
                    return True
                if event.key == KeyEvent.LeftKey:
                    self._moveView(-10,0)
                    return True

            if event.type == Event.MouseMoveEvent:

                if not self.getDragPlane():
                    if True:#cameraNotChanged
                        return True

                x = cast(MouseEvent, event).x
                y = cast(MouseEvent, event).y

                event = cast(MouseEvent, event)
                tool_id = self._selection_pass.getIdAtPosition(event.x, event.y)

                if not self.getDragStart():
                    self.setDragStart(x, y)
                    return False

                drag = self.getDragVector(x, y)
                if drag:
                    if not self._moved:
                        self._moved = True
                        self._distance = Vector(0, 0, 0)

                    #do sth here with the drag

                    if not self._distance:
                        self._distance = Vector(0, 0, 0)
                    self._distance += drag



                    if self._mode == Mode.MOVE:
                        #move the recthandle
                        self._handle.setPosition(self._handle.getPosition()+drag)
                        self._innerRectHandle.setPosition(self._innerRectHandle.getPosition()+drag)
                        self._leftCornerHandle.setPosition(self._leftCornerHandle.getPosition()+drag)
                        self._rightCornerHandle.setPosition(self._rightCornerHandle.getPosition()+drag)

                    elif self._mode == Mode.SCALE_LEFT_CORNER:
                        #resize the "leftcorner" handle
                        self._distance -= drag/2.0
                        drag *= 0.5

                        leftHandleCenterY = self._leftCornerHandle.getPosition().y+drag.y+1.25
                        rectCenterY = self._handle.getPosition().y

                        ydistance = abs(leftHandleCenterY - rectCenterY)
                        yscale = (ydistance / 1.25)

                        leftHandleCenterZ = self._leftCornerHandle.getPosition().z+drag.z-1.25
                        rectCenterZ = self._handle.getPosition().z

                        zdistance = abs(leftHandleCenterZ - rectCenterZ)
                        zscale = (zdistance/1.25)

                        cpos = self._handle.getPosition()
                        self._leftCornerHandle.setPosition(self._leftCornerHandle.getPosition()+drag*2)

                        self._handle.setPosition(Vector(cpos.x, cpos.y+drag.y, cpos.z+drag.z))
                        self._handle.setScale(Vector(1.0, yscale, zscale))

                        self._setInnerRectHandlePosAndSize()

                    elif self._mode == Mode.SCALE_RIGHT_CORNER:
                        #resize the "left corner" handle
                        self._distance -= drag/2.0
                        drag *= 0.5
                        #Logger.log("i", "right")
                        rightHandleCenterY = self._rightCornerHandle.getPosition().y+drag.y-1.25
                        rectCenterY = self._handle.getPosition().y

                        ydistance = abs(rightHandleCenterY - rectCenterY)
                        yscale = (ydistance / 1.25)

                        rightHandleCenterZ = self._rightCornerHandle.getPosition().z+drag.z+1.25
                        rectCenterZ = self._handle.getPosition().z

                        zdistance = abs(rightHandleCenterZ - rectCenterZ)
                        zscale = (zdistance/1.25)

                        cpos = self._handle.getPosition()
                        self._rightCornerHandle.setPosition(self._rightCornerHandle.getPosition()+drag*2)

                        self._handle.setPosition(Vector(cpos.x, cpos.y+drag.y, cpos.z+drag.z))
                        self._handle.setScale(Vector(1.0, yscale, zscale))

                        self._setInnerRectHandlePosAndSize()


                self.setDragStart(x, y)


            if event.type == Event.MousePressEvent and self._controller.getToolsEnabled():
            # Start a move/resize operation
                if MouseEvent.LeftButton not in cast(MouseEvent, event).buttons:
                    return False

                if not self._selection_pass:
                    return False
                id = self._selection_pass.getIdAtPosition(cast(MouseEvent, event).x, cast(MouseEvent, event).y)
                if not id:
                    return False

                if id == ToolHandle.XAxis:
                    self._mode = Mode.MOVE
                elif id == ToolHandle.YAxis:
                    self._mode = Mode.SCALE_LEFT_CORNER
                elif id == ToolHandle.ZAxis:
                    self._mode = Mode.SCALE_RIGHT_CORNER
                else:
                    return False

                self._moved = False

                camera = self._controller.getScene().getActiveCamera()
                if not camera:
                    return False
                camera_direction = camera.getPosition().normalized()

                #abs_x = abs(camera_direction.x)
                #abs_y = abs(camera_direction.y)

                Logger.log("i", "directions: %s %s %s", camera_direction.x,camera_direction.y, camera_direction.z)

                plane_vector = Vector(1.0, 0, 0).normalized()

                plane_distance = self._cutsurfacePosA.x if self._oneHalfSelected else self._cutsurfacePosB

                self.setDragPlane(Plane(plane_vector, plane_distance))
                return True

            if event.type == Event.MouseReleaseEvent:
            #finish a move/resize operation
                if self.getDragPlane():
                    self._distance = None
                    #self.propertyChanged.emit()
                    self.setDragPlane(None)
                    self.setDragStart(cast(MouseEvent, event).x, cast(MouseEvent, event).y)
                    self._isCubicConnectorValid()
                    self._adjustTolerance()

                    return True


        return False

    def _rootNodeChildrenChanged(self, node):

        if self._activeView == "PlaneSelection":
            #_intentionallyRemoved = true when the original node is removed
            if self._selectedObject not in node.getChildren() and not self._intentionallyRemoved:
                self._cancelPressed()
            if self._intentionallyRemoved:
                self._intentionallyRemoved = False
        if self._activeView == "CreateConnectors":
            if self._oneHalfNode not in node.getChildren() or self._otherHalfNode not in node.getChildren():
                #probably undo
                #finish in that case
                self._cancelPressed()

    def _selectedObjectTransformationChanged(self, object):
        Logger.log("i", "Not activated")
        #self._resizeCutsurfaceHandle()


    def _cancelPressed(self):
        """ called by UI
            cancels everything so far
            is also used to finish without creating connectors
        """

        if self._isSlicing:
            return
        #cancel everything here

        if self._activeView == "CreateConnectors":
            #set cameras back to normal
            self._removeFocusFromNode()
            if self._activeSubView == 1 or self._activeSubView == 3:
                Application.getInstance().getController().setCameraRotation("3d", 0)
            #remove created connectors
            self._removeAllCubicConnectors()

        self._removeCutsurfaceFromScene()
        self._removeHandles()

        self.setShowError(False)
        self.setErrorMessage("")
        self.setActiveView("Welcome")
        self.propertyChanged.emit()
        self._disableCutsurfaceHandle()

        self._areCubicConnectors3D = False
        self._cubicConnectors = []

    ################################
    ##
    # Welcome
    ##
    ################################

    def _startPressed(self):
        #called by ui
        #check if just one object is selected
        if Selection.getCount() == 1:
            self._oneHalfNode = None
            self._otherHalfNode = None
            self.setActiveView("PlaneSelection")
            self.propertyChanged.emit()
            #get selected node -
            self._selectedObject = Selection.getSelectedObject(0)

            #get center x position of slected object to position the plane
            self._sectionalPlaneX = self._selectedObject.getBoundingBox().center.x

            #detect if the transformation of the object changes
            #self._selectedObject.transformationChanged.connect(self._selectedObjectTransformationChanged)

            self._enableCutsurfaceHandle(0)


            scene = Application.getInstance().getController().getScene()
            rootNode = scene.getRoot()
            rootNode.childrenChanged.connect(self._rootNodeChildrenChanged)



    ################################
    ##
    # Cutsurface Selection
    ##
    ################################

    def _anErrorOccuredWhileCutting(self):
        self.setShowError(True)
        self.setErrorMessage("Something went wrong \nMesh is probably not watertight etc.")
        self.propertyChanged.emit()
        self._enableCutsurfaceHandle(1)

    def _finished(self):
        self._cancelPressed()

    def _selectPlanePressed(self):

        if self._isSlicing:
            return

        mesh = self._selectedObject.getMeshData().getTransformed(self._selectedObject.getLocalTransformation())
        job = Jobs.SliceMeshJob(mesh, [1.0, 0.0, 0.0], [self._sectionalPlaneX,0,0])
        job.finished.connect(self._onSliceMeshJobFinished)
        self.setIsSlicing(True)
        self.propertyChanged.emit()
        self._enableCutsurfaceHandle(2)
        job.start()

    def _onSliceMeshJobFinished(self, job):
        self.setIsSlicing(False)

        if job.getResult() is None:
            self._anErrorOccuredWhileCutting()
            return

        (oneHalfMesh, oneHalfCutSurfaceData), (otherHalfMesh, otherHalfCutSurfaceData) = job.getResult()

        self._oneHalfCutSurfaceData = oneHalfCutSurfaceData
        self._otherHalfCutSurfaceData = otherHalfCutSurfaceData

        active_build_plate = CuraApplication.getInstance().getMultiBuildPlateModel().activeBuildPlate

        originalName = self._selectedObject.getName()

        oneHalfNode = CuraSceneNode()
        oneHalfNode.transformationChanged.connect(self._transformChanged)
        oneHalfNode.setMeshData(oneHalfMesh)
        oneHalfNode.setSelectable(True)
        oneHalfNode.setName(originalName+"_A")
        oneHalfNode.addDecorator(BuildPlateDecorator(active_build_plate))
        oneHalfNode.addDecorator(SliceableObjectDecorator())

        otherHalfNode = CuraSceneNode()
        otherHalfNode.transformationChanged.connect(self._transformChanged)
        otherHalfNode.setMeshData(otherHalfMesh)
        otherHalfNode.setSelectable(True)
        otherHalfNode.setName(originalName+"_B")
        otherHalfNode.addDecorator(BuildPlateDecorator(active_build_plate))
        otherHalfNode.addDecorator(SliceableObjectDecorator())

        self._oneHalfNode = oneHalfNode
        self._otherHalfNode = otherHalfNode

        scene = Application.getInstance().getController().getScene()
        rootNode = scene.getRoot() # returns "root" SceneNode Object

        op = GroupedOperation()

        op.addOperation(AddSceneNodeOperation(oneHalfNode, rootNode))
        op.addOperation(AddSceneNodeOperation(otherHalfNode, rootNode))

        world_position = oneHalfNode.getWorldPosition()
        new_position = world_position.set(x = 30 + (world_position.x))
        self._oneHalfNodeYOffset = new_position.y
        node_op = TranslateOperation(oneHalfNode, new_position, set_position = True)
        op.addOperation(node_op)

        world_position = otherHalfNode.getWorldPosition()
        new_position = world_position.set(x = -30 + (world_position.x))
        self._otherHalfNodeYOffset = new_position.y
        node_op = TranslateOperation(otherHalfNode, new_position, set_position = True)
        op.addOperation(node_op)

        op.addOperation(RemoveSceneNodeOperation(self._selectedObject))
        self._intentionallyRemoved = True
        op.push()
        # go to step create connectors screen

        self._initCreateConnectorsView()

        """(openedMeshA, _, _) = oneHalfCutSurfaceData
        job = Jobs.CalculateAABBTreeJob(openedMeshA)
        job.finished.connect(self._onCalculateAABBTreeJobFinished)
        job.start()"""


    def _onCalculateAABBTreeJobFinished(self, job):
        if job.getResult() is None:
            Logger.log("i", "oh nooo :/")
        else:
            Logger.log("i", "woooorked!!")

    ## actions
    #

    def _disableCutsurfaceHandle(self):
        if not self._isHandleEnabled:
            return
        scene = Application.getInstance().getController().getScene()
        rootNode = scene.getRoot()
        rootNode.removeChild(self._sectionalPlaneToolHandle)
        self._isHandleEnabled = False

    # not yet in use

    def _resizeCutsurfaceHandle(self):
        if self._isHandleEnabled:
            node = self._sectionalPlaneToolHandle
            selectedObjectBoundingBox = self._selectedObject.getBoundingBox()
            planeBoundingBox = node.getBoundingBox()
            scale = max(selectedObjectBoundingBox.height, selectedObjectBoundingBox.depth) / planeBoundingBox.height
            node.setScale(Vector(1.0, scale*1.05, scale*1.05))
            node.setPosition(Vector(0.0, (selectedObjectBoundingBox.height/2.0)-(node.getPosition().y), 0.0))

    def _enableCutsurfaceHandle(self, style):
        if self._isHandleEnabled and style == self._planeStyle:
            return
        if not Selection.hasSelection():
            return

        self._planeStyle = style #
        self._selectedObject = Selection.getSelectedObject(0)


        #custom subclass of SceneNode
        node = PrettyNode.PrettyNode()
        node.setName("SectionalPlane")
        node.setSelectable(False)
        node.setSolidMesh(self._getCutsurfaceHandleMesh(self._planeStyle))


        selectedObjectBoundingBox = self._selectedObject.getBoundingBox()
        planeBoundingBox = node.getBoundingBox()
        #get max size depth /height
        scale = max(selectedObjectBoundingBox.height, selectedObjectBoundingBox.depth) / planeBoundingBox.height
        #get center on xy plane
        center = selectedObjectBoundingBox.center

        node.setScale(Vector(1.0, scale*1.05, scale*1.05))
        node.setPosition(Vector(self._sectionalPlaneX, (selectedObjectBoundingBox.height/2.0)-(node.getPosition().y), center.z))


        if self._isHandleEnabled:
            self._disableCutsurfaceHandle()

        scene = Application.getInstance().getController().getScene()
        rootNode = scene.getRoot() # returns "root" SceneNode Object
        self._sectionalPlaneToolHandle = node
        rootNode.addChild(self._sectionalPlaneToolHandle)
        self._isHandleEnabled = True


    def _getCutsurfaceHandleMesh(self, style) -> PrettyNode.PrettyNode:

        if self._sectionalPlaneMesh and style == 0:
            return self._sectionalPlaneMesh
        if self._sectionalPlaneAidMesh and style == 1:
            return self._sectionalPlaneAidMesh
        if self._sectionalPlaneCuttingMesh and style == 2:
            return self._sectionalPlaneCuttingMesh

        inner_radius = 40
        outer_radius = 40.5
        line_width = 0.5


        fillColor = Color(1.0, 0.0, 0.0, 0.5) if style == 0 else Color(0.0, 0.0, 0.0, 0.5)
        borderColor = Color(1.0, 0.0, 0.0, 1.0) if style == 0 else Color(0.0, 0.0, 0.0, 1.0)

        if style == 2:
            fillColor = Color(1.0, 0.0, 0.0, 0.5)
            borderColor = Color(1.0, 0.0, 0.0, 1.0)

        stripeColor = Color(1.0, 0.0, 0.0, 0.85)

        mb = MeshBuilder()
        mb.addDonut(
            inner_radius = 0.0,
            outer_radius = inner_radius,
            width = line_width,
            axis = Vector.Unit_Y,
            angle = math.pi / 2,
            color = fillColor
        )

        if style == 2:
            stripes = 2
            for i in range(0, math.floor(stripes/2)):

                stipe_width = inner_radius / stripes
                inner_r = (1+i*2)*stipe_width
                outer_r = inner_r + stipe_width

                mb.addDonut(
                    inner_radius = inner_r,
                    outer_radius = outer_r,
                    width = line_width,
                    axis = Vector.Unit_Y,
                    angle = math.pi / 2,
                    color = stripeColor
                )

        mb.addDonut(
            inner_radius = inner_radius,
            outer_radius = outer_radius,
            width = line_width,
            axis = Vector.Unit_Y,
            angle = math.pi / 2,
            color = borderColor
        )

        if style == 0:
            self._sectionalPlaneMesh = mb.build()
            return self._sectionalPlaneMesh
        if style == 1:
            self._sectionalPlaneAidMesh = mb.build()
            return self._sectionalPlaneAidMesh
        else:
            self._sectionalPlaneCuttingMesh = mb.build()
            return self._sectionalPlaneCuttingMesh


    ################################
    ##
    # Create Connectors
    ##
    ################################


    def _initCreateConnectorsView(self):

        """CreateConnectors: set up Window 1 """

        self._cutSurfaceCenterPosition = None
        self._rectHandleWithinPoly = False
        self._rectHandleCollidesWithOtherCubicConnectors = True

        self.setErrorMessage("")
        self.setActiveView("CreateConnectors")
        self.setActiveSubView(0)

        Selection.clear()
        Selection.add(self._oneHalfNode)

        self._disableCutsurfaceHandle()
        #clear connectorsList
        self._connectorsList.setItems([])
        self.propertyChanged.emit()

        # maybe better in a thread

        (_, cutSurfaceMesh, _) = self._oneHalfCutSurfaceData
        #we must convert it to MeshData
        cutSurfaceMesh = Operations.toMeshData(cutSurfaceMesh)
        #colorize pink
        cutSurfaceMesh = Operations.colorizeMesh(cutSurfaceMesh, Color(0.0, 0.0, 0.0, 1.0))

        # till here

        node = PrettyNode.PrettyNode()
        node.setName("CutSurface")
        node.setSelectable(False)
        node.setSolidMesh(cutSurfaceMesh)
        self._cutsurfaceNode = node

    def _createConnectorsPressed(self):

        """ CreateConnectors: Set up Widnow Two - called by UI """
        self.setActiveSubView(1)
        self.propertyChanged.emit()
        self._addCutsurfaceToScene()
        self._switchToCutSurface(True)

    ###
    ## Functions add or remove the cut surface mesh (to darken the surface)
    ###

    def _addCutsurfaceToScene(self):

        node = self._cutsurfaceNode
        scene = Application.getInstance().getController().getScene()
        rootNode = scene.getRoot()
        rootNode.addChild(node)

        op = GroupedOperation()
        world_position = node.getWorldPosition()
        new_position = world_position.set(x = 30 + (world_position.x), y = self._oneHalfNodeYOffset)
        node_op = TranslateOperation(node, new_position, set_position = True)
        op.addOperation(node_op)
        op.push()

    def _removeCutsurfaceFromScene(self):
        if self._cutsurfaceNode:
            self._cutsurfaceNode.setParent(None)


    ###
    ## Functions to change the view/perspective
    ###

    def _getCurrentHandleDepth(self):

        if self._oneHalfSelected:
            if self._connectorGender == 0 or self._connectorGender == 1:
                return self._depthA
            else:
                return self._depthB
        else:
            if self._connectorGender == 0 or self._connectorGender == 2:
                return self._depthB
            else:
                return self._depthA

    def _getCameraPositionAndLookAtFor(self, cview = CamView.FRONTAL):
        """ returns the camera position and lookAtPosition for a specific perspective """
        xpos = self._cutsurfacePosA.x if self._oneHalfSelected else self._cutsurfacePosB.x
        cnode = self._oneHalfNode if self._oneHalfSelected else self._otherHalfNode
        boundingBox = cnode.getBoundingBox()
        center = boundingBox.center

        if cview == CamView.FRONTAL:
            distance = max(boundingBox.height, boundingBox.width)
            return Vector(center.x-boundingBox.width-distance*2,center.y,center.z), None

        if cview == CamView.TOP:
            distance = max(boundingBox.height, boundingBox.width)
            return Vector(xpos,center.y+boundingBox.height*2+boundingBox.width+self._handle.getScale().z*2.5,self._handle.getWorldPosition().z), Vector(xpos,-1,self._handle.getWorldPosition().z)

        elif cview== CamView.BOTTOM:
            distance = max(boundingBox.depth, boundingBox.width)
            return Vector(xpos,-distance*2,self._handle.getWorldPosition().z), Vector(xpos,0,self._handle.getWorldPosition().z)

        elif cview == CamView.LEFT:
            direction = 1 if self._oneHalfSelected else -1
            distance = max(boundingBox.depth, boundingBox.height)
            return Vector(xpos,self._handle.getWorldPosition().y, -distance*2*direction), Vector(xpos,self._handle.getWorldPosition().y, 0)

        elif cview == CamView.RIGHT:
            direction = 1 if self._oneHalfSelected else -1
            distance = max(boundingBox.depth, boundingBox.height)
            return Vector(xpos,self._handle.getWorldPosition().y, distance*2*direction), Vector(xpos,self._handle.getWorldPosition().y, 0)



    def _switchToCutSurfaceA(self):
        """ called by UI """
        self._switchToCutSurface(True)
    def _switchToCutSurfaceB(self):
        """ called by UI"""
        self._switchToCutSurface(False)

    def _switchToCutSurface(self, oneHalf):
        """  """
        self._notFrontalViewActive = False
        self._cameraView = CamView.FRONTAL.value
        self._viewPosX = 0
        self._viewPosY = 0

        controller = Application.getInstance().getController()
        cameraTool = controller.getCameraTool()
        scene = controller.getScene()

        scene.getRoot().addChild(self._cutsurfaceNode)
        #centerOfCutsurface = self._getCenterOfCutsurface()

        if oneHalf:


            #get center x position of slected object to position the plane
            boundingBox = self._oneHalfNode.getBoundingBox()
            center = boundingBox.center
            #focus on _oneHalfNode (hide, disable and set unselectable all others )
            self._focusOnNode(self._oneHalfNode)
            # set camera position and rotation on cut surface
            #controller.setCameraRotation("y", 0)
            controller.setCameraRotation("x", 90)
            #cameraTool.setOrigin(Vector(center.x+boundingBox.width+distance*2,center.y,center.z))
            #calculate camera distance
            distance = max(boundingBox.height, boundingBox.depth)

            controller.setCameraPosition(center.x-distance*1.75,center.y,center.z)

            self._oneHalfSelected = True


            if(self._cutsurfaceNode):
                wpos = self._cutsurfaceNode.getWorldPosition()
                new_position = wpos.set(x=self._oneHalfNode.getWorldPosition().x, y = self._oneHalfNodeYOffset)
                self._cutsurfaceNode.setPosition(new_position, SceneNode.TransformSpace.World)

                if(self._activeSubView == 2): #adding a new connector is active
                    self._setHandleDepth(0.01)
                    self._moveHandles(True)

        if not oneHalf:


            #get center x position of selected object to position the plane
            boundingBox = self._otherHalfNode.getBoundingBox()
            center = boundingBox.center
            #focus on _otherHalfNode (hide, disable and set unselectable all others )
            self._focusOnNode(self._otherHalfNode)

            #calculate camera distance
            distance = max(boundingBox.height, boundingBox.depth)

            cameraTool = controller.getCameraTool()
            # set camera position and rotation on cut surface

            camera = controller.getScene().getActiveCamera()

            #controller.setCameraRotation("y", 0)
            controller.setCameraRotation("x", -90)
            #cameraTool.setOrigin(Vector(center.x+boundingBox.width+distance*2,center.y,center.z))

            controller.setCameraPosition(center.x+distance*1.75,center.y,center.z)

            self._oneHalfSelected = False

            if(self._cutsurfaceNode):
                wpos = self._cutsurfaceNode.getWorldPosition()
                new_position = wpos.set(x=self._otherHalfNode.getWorldPosition().x, y = self._otherHalfNodeYOffset)
                self._cutsurfaceNode.setPosition(new_position, SceneNode.TransformSpace.World)

                if(self._activeSubView == 2): #adding a new connector is active
                    self._setHandleDepth(0.01)
                    self._moveHandles(True)
        self._setVisibilityOfCubicConnectors(True)
        self.propertyChanged.emit()


    def _switchToBirdViewA(self):
        change = False
        if not self._oneHalfSelected:
            change = True
            self._oneHalfSelected = True
            self._focusOnNode(self._oneHalfNode)
        if self._cameraView == CamView.FRONTAL.value or change:
            self._cameraView = CamView.TOP.value
            self._moveHandles(False)
            self._switchView()
            self.propertyChanged.emit()

    def _switchToBirdViewB(self):
        change = False
        if self._oneHalfSelected:
            self._oneHalfSelected = False
            self._focusOnNode(self._otherHalfNode)
            change = True

        if self._cameraView == CamView.FRONTAL.value or change:
            self._cameraView = CamView.TOP.value
            self._moveHandles(False)
            self._switchView()
            self.propertyChanged.emit()

    def _switchView(self):
        """ changes the view to the prior set self._cameraView """
        oneHalf = self._oneHalfSelected
        self._notFrontalViewActive = True

        partNode = self._oneHalfNode if self._oneHalfSelected else self._otherHalfNode
        clearNode = self._oneHalfNode if not self._oneHalfSelected else self._otherHalfNode

        self._viewPosX = 0
        self._viewPosY = 0

        controller = Application.getInstance().getController()
        cameraTool = controller.getCameraTool()
        scene = controller.getScene()

        #remove handles and cut surface
        scene.getRoot().removeChild(self._cutsurfaceNode)

        #self._handle.setParent(None)
        self._innerRectHandle.setParent(None)
        self._leftCornerHandle.setParent(None)
        self._rightCornerHandle.setParent(None)


        self._setHandleDepth(self._getCurrentHandleDepth())
        self._moveHandles(False)
        self._setVisibilityOfCubicConnectors(False)

        if self._cameraView == CamView.TOP.value:
            controller.setCameraRotation("x", 0)
            controller.setCameraRotation("y", 90)
            pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.TOP)
            controller.setCameraPosition(pos.x, pos.y, pos.z)
            #controller.setLookAtPosition(lookAt.x, lookAt.y, lookAt.z)


        elif self._cameraView == CamView.BOTTOM.value:
            controller.setCameraRotation("x", 0)
            controller.setCameraRotation("y", 90)
            pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.BOTTOM)
            controller.setCameraPosition(pos.x, pos.y, pos.z)
            controller.setLookAtPosition(lookAt.x, lookAt.y, lookAt.z)


        elif self._cameraView == CamView.LEFT.value:
            controller.setCameraRotation("x", 0)
            controller.setCameraRotation("y", 0)
            pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.LEFT)
            controller.setCameraPosition(pos.x, pos.y, pos.z)
            controller.setLookAtPosition(lookAt.x, lookAt.y, lookAt.z)


        elif self._cameraView == CamView.RIGHT.value:
            controller.setCameraRotation("x", 0)
            controller.setCameraRotation("y", 0)
            pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.RIGHT)
            controller.setCameraPosition(pos.x, pos.y, pos.z)
            controller.setLookAtPosition(lookAt.x, lookAt.y, lookAt.z)

    def _moveView(self, x, y):
        """
            moves the whole view left, right, up, down in relation to the frontal cutsurface
        """
        controller = Application.getInstance().getController()
        camera = controller.getScene().getActiveCamera()

        if self._oneHalfSelected:

            p = 1.0

            boundingBox = self._oneHalfNode.getBoundingBox()
            center = boundingBox.center
            distance = max(boundingBox.height, boundingBox.width)

            if camera.getPosition().x < 0:
                p = (camera.getPosition().x / (center.x+boundingBox.width+distance*2))
            else:
                p = 0.1
            p = abs(p)
            if p < 0.1:
                p = 0.1
            self._viewPosX = self._viewPosX + x*p
            self._viewPosY = self._viewPosY + y*p
            if(self._notFrontalViewActive):

                if self._cameraView == CamView.TOP.value:
                    pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.TOP)
                    controller.setCameraPosition(pos.x+self._viewPosX, camera.getPosition().y, pos.z-self._viewPosY)

                if self._cameraView == CamView.BOTTOM.value:
                    pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.BOTTOM)
                    controller.setCameraPosition(pos.x+self._viewPosX, camera.getPosition().y, pos.z+self._viewPosY)

                if self._cameraView == CamView.LEFT.value:
                    pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.LEFT)
                    controller.setCameraPosition(pos.x-self._viewPosX, pos.y+self._viewPosY, camera.getPosition().z)

                if self._cameraView == CamView.RIGHT.value:
                    pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.RIGHT)
                    controller.setCameraPosition(pos.x+self._viewPosX, pos.y+self._viewPosY, camera.getPosition().z)

                #controller.setCameraPosition(self._oneHalfNode.getWorldPosition().x+self._viewPosX,camera.getPosition().y,center.z-self._viewPosY)
            else:
                controller.setCameraPosition(camera.getPosition().x,center.y+self._viewPosY,center.z+self._viewPosX)
        else:
            p = 1.0
            boundingBox = self._otherHalfNode.getBoundingBox()
            center = boundingBox.center
            distance = max(boundingBox.height, boundingBox.width)

            if camera.getPosition().x > 0:
                p = (camera.getPosition().x / (center.x+boundingBox.width+distance*2))
            else:
                p = 0.1
            p = abs(p)
            if p < 0.1:
                p = 0.1

            self._viewPosX = self._viewPosX + x*p*(-1)
            self._viewPosY = self._viewPosY + y*p
            if(self._notFrontalViewActive):
                if self._cameraView == CamView.TOP.value:
                    pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.TOP)
                    controller.setCameraPosition(pos.x-self._viewPosX, camera.getPosition().y, pos.z-self._viewPosY)

                if self._cameraView == CamView.BOTTOM.value:
                    pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.BOTTOM)
                    controller.setCameraPosition(pos.x+self._viewPosX, camera.getPosition().y, pos.z-self._viewPosY)

                if self._cameraView == CamView.LEFT.value:
                    pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.LEFT)
                    controller.setCameraPosition(pos.x-self._viewPosX, pos.y+self._viewPosY, camera.getPosition().z)

                if self._cameraView == CamView.RIGHT.value:
                    pos, lookAt = self._getCameraPositionAndLookAtFor(cview=CamView.RIGHT)
                    controller.setCameraPosition(pos.x+self._viewPosX, pos.y+self._viewPosY, camera.getPosition().z)
            else:
                controller.setCameraPosition(camera.getPosition().x,center.y+self._viewPosY,center.z+self._viewPosX)


    ###
    ## Functions related to the rect-handle
    ###

    def _transformChanged(self, node : SceneNode):
        """
            detects if oneHalfNode or otherHalfNode falls onto the buildplate
            and saves the offset to adjust the rect-handle
        """
        if node == self._oneHalfNode:
            world_position = node.getWorldPosition()
            if world_position.y != self._oneHalfNodeYOffset:
                self._oneHalfNodeYOffset = world_position.y
                # move tooltips if oneHalfSelected
                if self._oneHalfSelected:
                    wpos = self._cutsurfaceNode.getWorldPosition()
                    new_position = wpos.set(y = self._oneHalfNodeYOffset)
                    self._cutsurfaceNode.setPosition(new_position, SceneNode.TransformSpace.World)

        if node == self._otherHalfNode:
            world_position = node.getWorldPosition()
            if world_position.y != self._otherHalfNodeYOffset:
                self._otherHalfNodeYOffset = world_position.y
                # move tooltips if !oneHalfSelected
                if not self._oneHalfSelected:
                    wpos = self._cutsurfaceNode.getWorldPosition()
                    new_position = wpos.set(y = self._otherHalfNodeYOffset)
                    self._cutsurfaceNode.setPosition(new_position, SceneNode.TransformSpace.World)



    def _getCenterOfCutsurface(self):
        """ returns the center of the cutsurface """
        if not self._cutSurfaceCenterPosition is None:
            return self._cutSurfaceCenterPosition

        from shapely.geometry import Point
        from trimesh import transformations
        (_, _, onPlaneAnd3D) = self._oneHalfCutSurfaceData
        (on_plane, to3D) = onPlaneAnd3D
        center_points = []

        x_min, x_max = 0, 0
        y_min, y_max = 0, 0
        first = True
        for polygon in on_plane.polygons_closed:
            (x1, y1, x2, y2) = polygon.bounds
            if x1 < x_min or first:
                x_min = x1
            if y1 < y_min or first:
                y_min = y1
            if x2 > x_max or first:
                x_max = x2
            if y2 > y_max or first:
                y_max = y2
            if first:
                first = False

        center_points.append([x_min + ((x_max - x_min) / 2), y_min + ((y_max - y_min) / 2), 0]);
        center_points = transformations.transform_points(center_points, to3D)
        self._cutSurfaceCenterPosition = Vector(center_points[0][0], center_points[0][1], center_points[0][2])
        return Vector(center_points[0][0], center_points[0][1], center_points[0][2])

    def _getCenterOfBiggestCutsurfacePolygon(self) -> Vector:
        """ returns the center of the biggest polygon of the cutsurface """
        from shapely.geometry import Point
        from trimesh import transformations

        (_, _, onPlaneAnd3D) = self._oneHalfCutSurfaceData
        (on_plane, to3D) = onPlaneAnd3D

        center_points = []
        center_areas = []

        for polygon in on_plane.polygons_closed:
            (x1, y1, x2, y2) = polygon.bounds
            center_areas.append(polygon.area)
            center_points.append([x1 + ((x2 - x1) / 2), y1 + ((y2 - y1) / 2), 0]);

        center_points = transformations.transform_points(center_points, to3D)
        max_i = center_areas.index(max(center_areas))

        return Vector(center_points[max_i][0], center_points[max_i][1], center_points[max_i][2])

    def _removeHandles(self):
        if self._handle:
            self._handle.setParent(None)
            self._handle.setEnabled(False)
        if self._innerRectHandle:
            self._innerRectHandle.setParent(None)
            self._innerRectHandle.setEnabled(False)
        if self._leftCornerHandle:
            self._leftCornerHandle.setParent(None)
            self._leftCornerHandle.setEnabled(False)
        if self._rightCornerHandle:
            self._rightCornerHandle.setParent(None)
            self._rightCornerHandle.setEnabled(False)


    def _createHandles(self):
        #Logger.log("i", "once")
        """
        creates the handles (leftcornerhandle, rightcornerhandle, rect, innerrect)
        """

        self._handle.setParent(self.getController().getScene().getRoot())
        self._handle.setEnabled(True)

        handleScale = Vector(1.0, 1.0, 1.0)
        handlePosition = None
        leftCornerHandlePosition = None
        rightCornerHandlePosition = None

        if self._isEditingCubicConnector:
            (nodea, nodeb, _, _, _, _,_) = self._cubicConnectors[self.getSelectedConnectorIndex()]
            handleScale = Vector(1.0, nodea.getScale().y, nodea.getScale().z)
            handlePosition = Vector(self._sectionalPlaneToolHandle.getWorldPosition().x+30 ,nodea.getWorldPosition().y, nodea.getWorldPosition().z) if self._oneHalfSelected else Vector(self._sectionalPlaneToolHandle.getWorldPosition().x-30 ,nodeb.getWorldPosition().y, nodea.getWorldPosition().z)

            ww = -(nodea.getScale().z*2.5)/2.0 + 1.25
            hh = (nodea.getScale().y*2.5)/2.0 - 1.25

            leftCornerHandlePosition = Vector(handlePosition.x, handlePosition.y+hh, handlePosition.z+ww)
            rightCornerHandlePosition = Vector(handlePosition.x, handlePosition.y-hh, handlePosition.z-ww)


        self._handle.setScale(handleScale)

        handle_y_scale = self._handle.getScale().y
        handle_z_scale = self._handle.getScale().z
        innerWidth = (self._handle.getScale().z*2.5)-self._innerHandleBorder
        innerHeight = (self._handle.getScale().y*2.5)-self._innerHandleBorder
        innerZScale = innerWidth / 2.5
        innerYScale = innerHeight / 2.5

        self._innerRectHandle.setParent(self.getController().getScene().getRoot())
        self._innerRectHandle.setEnabled(True)

        self._innerRectHandle.setScale(Vector(self._handle.getScale().x, innerYScale, innerZScale))

        self._leftCornerHandle.setParent(self.getController().getScene().getRoot())
        self._leftCornerHandle.setEnabled(True)
        self._leftCornerHandle.setScale(Vector(1.0, 1.0, 1.0))

        self._rightCornerHandle.setParent(self.getController().getScene().getRoot())
        self._rightCornerHandle.setEnabled(True)
        self._rightCornerHandle.setScale(Vector(1.0, 1.0, 1.0))

        self._cutsurfacePosA = Vector(self._sectionalPlaneToolHandle.getWorldPosition().x+30, 0, self._sectionalPlaneToolHandle.getWorldPosition().z)
        self._cutsurfacePosB = Vector(self._sectionalPlaneToolHandle.getWorldPosition().x-30, 0, self._sectionalPlaneToolHandle.getWorldPosition().z)


        if self._oneHalfSelected:
            #change to: get poly center + y offset
            startPosition = self._getCenterOfBiggestCutsurfacePolygon()
            startPosition = Vector(self._cutsurfacePosA.x, startPosition.y+self._oneHalfNodeYOffset,startPosition.z)
            Logger.log("i", "Possible Center: %s", startPosition)
            handlePosition = startPosition if handlePosition is None else handlePosition

            self._currentHandleYOffset = self._oneHalfNodeYOffset
        else:
            #change to: get poly center + y offset
            startPosition = self._getCenterOfBiggestCutsurfacePolygon()
            startPosition = Vector(self._cutsurfacePosB.x, startPosition.y+self._otherHalfNodeYOffset,startPosition.z)
            Logger.log("i", "Possible Center: %s", startPosition)

            handlePosition = startPosition if handlePosition is None else handlePosition
            self._currentHandleYOffset = self._otherHalfNodeYOffset

        leftCornerHandlePosition = leftCornerHandlePosition if leftCornerHandlePosition is not None else handlePosition
        rightCornerHandlePosition = rightCornerHandlePosition if rightCornerHandlePosition is not None else handlePosition

        self._setHandleGender()

        self._handle.setPosition(handlePosition, SceneNode.TransformSpace.World)
        self._innerRectHandle.setPosition(handlePosition, SceneNode.TransformSpace.World)

        self._leftCornerHandle.setPosition(leftCornerHandlePosition, SceneNode.TransformSpace.World)
        self._rightCornerHandle.setPosition(rightCornerHandlePosition, SceneNode.TransformSpace.World)

        self._isCubicConnectorValid()


    def _setHandleGender(self):
        """ changes the color of the handle
            red: Female
            green: Male
            black: invalid position / size / depth
        """
        h_disabled = not self._rectHandleWithinPoly or self._rectHandleCollidesWithOtherCubicConnectors

        if self._connectorGender == 0:
            #both female
            self._handle.changeHandleMode(female=True, disabled=h_disabled) # true equals female
            self._innerRectHandle.changeHandleMode(female=True, disabled=h_disabled)
        elif self._connectorGender == 1:
            #onehalf (a) is female / otherhalf (b) is male
            if self._oneHalfSelected:
                self._handle.changeHandleMode(female=True, disabled=h_disabled)
                self._innerRectHandle.changeHandleMode(female=True, disabled=h_disabled)
            else:
                self._handle.changeHandleMode(female=False, disabled=h_disabled)
                self._innerRectHandle.changeHandleMode(female=False, disabled=h_disabled)
        else:
            #onehalf (a) is male / ohterhalf (b) is female
            if self._oneHalfSelected:
                self._handle.changeHandleMode(female=False, disabled=h_disabled)
                self._innerRectHandle.changeHandleMode(female=False, disabled=h_disabled)
            else:
                self._handle.changeHandleMode(female=True, disabled=h_disabled)
                self._innerRectHandle.changeHandleMode(female=True, disabled=h_disabled)


    def _setHandleDepth(self, new):
        """ mostly used to make the handle flat for the frontal view and 3d vice versa """

        self._handle.setScale(Vector(new/0.5, self._handle.getScale().y, self._handle.getScale().z))
        self._currentDepth = new

    def _adjustTolerance(self):
        minimum = min(self._handle.getScale().z*2.5, self._handle.getScale().y*2.5)/5.0
        if self._tolerance >= minimum:
            self.setTolerance(round(minimum, 2))
            #Logger.log("i", "new tolerance: %s", self._tolerance)
            self.propertyChanged.emit()

    def _setInnerRectHandlePosAndSize(self):
        handle_y_scale = self._handle.getScale().y
        handle_z_scale = self._handle.getScale().z
        innerWidth = (self._handle.getScale().z*2.5)-self._innerHandleBorder
        innerHeight = (self._handle.getScale().y*2.5)-self._innerHandleBorder
        innerZScale = innerWidth / 2.5
        innerYScale = innerHeight / 2.5
        self._innerRectHandle.setPosition(Vector(self._handle.getWorldPosition().x, self._handle.getWorldPosition().y, self._handle.getWorldPosition().z), SceneNode.TransformSpace.World)
        self._innerRectHandle.setScale(Vector(self._handle.getScale().x, innerYScale, innerZScale))

    def _moveHandles(self, enable):
        """ moves the handle from half a to half b and vice versa
            enable = True - adds all handles to the view
        """

        if self._cutsurfacePosA is None or self._cutsurfacePosB is None:
            return

        depth = self._currentDepth
        xPosCutsurface = None

        if self._oneHalfSelected:
            node = self._oneHalfNode
            depth = self._depthA if self._connectorGender == 0 or self._connectorGender == 1 else self._depthB
            xPosCutsurface = self._cutsurfacePosA.x
        else:
            node = self._otherHalfNode
            depth = -self._depthB if self._connectorGender == 0 or self._connectorGender == 2 else -self._depthA
            xPosCutsurface = self._cutsurfacePosB.x


        oldHandleYOffset = self._currentHandleYOffset
        self._currentHandleYOffset = self._oneHalfNodeYOffset if self._oneHalfSelected else self._otherHalfNodeYOffset
        offsetDifference = self._currentHandleYOffset - oldHandleYOffset


        if enable:
            self._handle.setParent(None)
            self._handle.setParent(self.getController().getScene().getRoot())
            self._innerRectHandle.setParent(self.getController().getScene().getRoot())
            self._leftCornerHandle.setParent(self.getController().getScene().getRoot())
            self._rightCornerHandle.setParent(self.getController().getScene().getRoot())


        self._setHandleGender()

        hDirection = 1 if self._connectorGender == 0 or (self._oneHalfSelected and self._connectorGender == 1) or (not self._oneHalfSelected and self._connectorGender == 2) else -1

        #just in the case they must be flat
        if self._activeView == "CreateConnectors" and self._activeSubView == 2 and not self._notFrontalViewActive:
            hDirection = 0

        xpos = xPosCutsurface+(depth/2.0)*hDirection


        self._handle.setPosition(Vector(xpos, self._handle.getWorldPosition().y+offsetDifference, self._handle.getWorldPosition().z), SceneNode.TransformSpace.World)
        self._setInnerRectHandlePosAndSize()


        self._leftCornerHandle.setPosition(Vector(xpos, self._leftCornerHandle.getWorldPosition().y+offsetDifference, self._leftCornerHandle.getWorldPosition().z), SceneNode.TransformSpace.World)
        self._rightCornerHandle.setPosition(Vector(xpos, self._rightCornerHandle.getWorldPosition().y+offsetDifference, self._rightCornerHandle.getWorldPosition().z), SceneNode.TransformSpace.World)

    def _isCubicConnectorValid(self):
        """
            checks if the new connectors is inside the polygon, is not colliding with others

            (checked also before whether the connector collides with the mesh (to deep!)
            removed due to perfomance issues and instead it's  being checked once before adding it finally)
        """

        self._isRectWithinPoly()
        self._isCollidingWithOtherCubicConnectos()

        #self._isCollidingWithMesh()

    def _isCollidingWithOtherCubicConnectos(self):
        """
            checks if the handle collides with already created cubic connectors
        """
        if not self._isCollidingWithOtherCCJob is None:
            if self._isCollidingWithOtherCCJob.isRunning():
                self._isCollidingWithOtherCCJob.finished.disconnect(self._onIsCollidingWithOtherCCJobFinished)

        cubicConnectors = self._cubicConnectors if not self._isEditingCubicConnector else [x for i,x in enumerate(self._cubicConnectors) if i!=self.getSelectedConnectorIndex()]

        self._isCollidingWithOtherCCJob = Jobs.IsCollidingWithOtherRectsJob(self._handle, cubicConnectors, self._oneHalfSelected)
        self._isCollidingWithOtherCCJob.finished.connect(self._onIsCollidingWithOtherCCJobFinished)
        self._isCollidingWithOtherCCJob.start()

    def _onIsCollidingWithOtherCCJobFinished(self, job):
        if job.getResult() is None:
            return
        collide = job.getResult()
        if self._rectHandleCollidesWithOtherCubicConnectors != collide:
            self._rectHandleCollidesWithOtherCubicConnectors = collide
            self._setHandleGender()
            self.propertyChanged.emit()

    def _isCollidingWithMesh(self):
        """
            checks if the new cube would collide with the mesh (limits the depth of the new connectors)
        """


        if not self._isCollidingWithMeshJob is None:
            if self._isCollidingWithMeshJob.isRunning():
                self._isCollidingWithMeshJob.finished.disconnect(self._onIsCollidingWithMeshFinished)

        hasToCheckA = self._connectorGender == 0 or self._connectorGender == 1
        hasToCheckB = self._connectorGender == 0 or self._connectorGender == 2
        hOffset = self._oneHalfNodeYOffset if self._oneHalfSelected else self._otherHalfNodeYOffset

        self._isCollidingWithMeshJob = Jobs.IsCollidingWithMeshJob(self._oneHalfCutSurfaceData, self._otherHalfCutSurfaceData, Vector(self._sectionalPlaneX, self._handle.getWorldPosition().y-hOffset, self._handle.getWorldPosition().z), self._handle.getScale(), self._depthA, self._depthB, hasToCheckA, hasToCheckB)
        self._isCollidingWithMeshJob.finished.connect(self._onIsCollidingWithMeshFinished)
        self._isCollidingWithMeshJob.start()


    def _onIsCollidingWithMeshFinished(self, job):
        if job.getResult() is None:
            return
        (collidesWithOneHalf, collidesWithOtherHalf) = job.getResult()

        if collidesWithOneHalf or collidesWithOtherHalf:
            self._checksCollidingWithMesh = False
            self._activeSubView = 2

            bgng = "Cubic connector goes too far. \n Change \"Depth "
            self._errorMsg = bgng + "A / B\"." if collidesWithOneHalf and collidesWithOtherHalf else (bgng + "A\"." if collidesWithOneHalf else bgng + "B\".")
            self.setShowDepthError(True)

            self.propertyChanged.emit()
        else:
            self._addCubicConnector()



    def _isRectWithinPoly(self):
        """
            checks if the rect is within one polygon of the cutsurface
        """

        if not self._isRectInPolyJob is None:
            if self._isRectInPolyJob.isRunning():
                self._isRectInPolyJob.finished.disconnect(self._onIsRectWithinPolyJobFinished)

        currYOffset = self._oneHalfNodeYOffset if self._oneHalfSelected else self._otherHalfNodeYOffset
        rectData = (Vector(self._sectionalPlaneX, self._handle.getWorldPosition().y-currYOffset, self._handle.getWorldPosition().z), self._handle.getScale().z*2.5, self._handle.getScale().y*2.5)

        self._isRectInPolyJob = Jobs.IsRectInPolyJob(rectData, self._oneHalfCutSurfaceData)
        self._isRectInPolyJob.finished.connect(self._onIsRectWithinPolyJobFinished)
        self._isRectInPolyJob.start()


    def _onIsRectWithinPolyJobFinished(self, job):

        if job.getResult() is None:
            return
        if self._rectHandleWithinPoly !=  job.getResult():
            self._rectHandleWithinPoly = job.getResult()
            self._setHandleGender()
            self.propertyChanged.emit()


        if job.getResult():
            Logger.log("i", "is in poly")
        else:
            Logger.log("i", "it's not")


    ###
    ## Functions to show or hide all nodes except one
    ###


    def _focusOnNode(self, focusNode):
        """ hides all nodes on the buildplate except the chosen one """
        for node in DepthFirstIterator(Application.getInstance().getController().getScene().getRoot()):
            if node.hasDecoration("isSliceable") or node.hasDecoration("isBlockSlicing") or node.hasDecoration("setBuildPlateNumber") or node.hasDecoration("getAdhesionArea"):
                if node != focusNode:
                    """if node not in (self._temporarilyRemovedNodes):
                        node.setParent(None)
                        self._temporarilyRemovedNodes.append(node)"""

                    if node.isVisible():
                        node.setVisible(False)
                        self._hiddenNodes.append(node)
                    if node.isEnabled():
                        node.setEnabled(False)
                        self._disabledNodes.append(node)
                    if node.isSelectable():
                        node.setSelectable(False)
                        self._unselectedNodes.append(node)


        if focusNode in self._hiddenNodes:
            focusNode.setVisible(True)
            self._hiddenNodes.remove(focusNode)
        if focusNode in self._disabledNodes:
            focusNode.setEnabled(True)
            self._disabledNodes.remove(focusNode)
        if focusNode in self._unselectedNodes:
            focusNode.setSelectable(True)
            self._unselectedNodes.remove(focusNode)

        Selection.clear()
        Selection.add(focusNode)


    def _removeFocusFromNode(self):
        """ reverse _focusOnNode """
        #Logger.log("i", "nodesHidden: %s", self._hiddenNodes)
        for i in range(0, len(self._hiddenNodes)):
            self._hiddenNodes[i].setVisible(True)
        for i in range (0, len(self._disabledNodes)):
            self._disabledNodes[i].setEnabled(True)
        for i in range (0, len(self._unselectedNodes)):
            self._unselectedNodes[i].setSelectable(True)

        self._hiddenNodes = []
        self._disabledNodes = []
        self._unselectedNodes = []

        self._oneHalfSelected = False


    ###
    ## Functions to add or remove cubic connectors
    ##


    def _beginAddingCubicConnector(self):
        self.setShowDepthError(False)
        self._isEditingCubicConnector = False
        self._tolerance = 0.1
        self._rectHandleCollidesWithOtherCubicConnectors = True
        self._rectHandleWithinPoly = False
        self._activeSubView = 2

        self._createHandles()
        self.propertyChanged.emit()
        self._toggleCubicConnectors3DAndPlanar(to3D = False)

    def _editCubicConnector(self):
        if not (self.getSelectedConnectorIndex() >= 0 and self.getSelectedConnectorIndex() < len(self._cubicConnectors)):
            return
        self._isEditingCubicConnector = True
        self.setShowDepthError(False)
        (nodea, nodeb, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = self._cubicConnectors[self.getSelectedConnectorIndex()]
        nodea.setParent(None)
        nodeb.setParent(None)



        self._tolerance = addtnlProps["tolerance"]
        self._depthA = depthA
        self._depthB = depthB
        self._connectorGender = 0 if isFemaleA and isFemaleB else (1 if isFemaleA else 2)


        self._isEditingCubicConnector = True
        self._rectHandleCollidesWithOtherCubicConnectors = True
        self._rectHandleWithinPoly = False
        self._activeSubView = 2
        self._createHandles()
        self.propertyChanged.emit()
        self._toggleCubicConnectors3DAndPlanar(to3D = False)

    def _cancelAddingConnector(self):
        self.setShowDepthError(False)
        if self._isEditingCubicConnector:
            (nodea, nodeb, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = self._cubicConnectors[self.getSelectedConnectorIndex()]
            rootNode = Application.getInstance().getController().getScene().getRoot()
            if self._oneHalfSelected:
                nodea.setParent(rootNode)
            else:
                nodeb.setParent(rootNode)

        self._isEditingCubicConnector = False
        self._rectHandleCollidesWithOtherCubicConnectors = True
        self._rectHandleWithinPoly = False
        self._activeSubView = 1
        self._removeHandles()
        self.propertyChanged.emit()
        self._toggleCubicConnectors3DAndPlanar(to3D = True)

        if self._cameraView != CamView.FRONTAL.value:
            self._switchToCutSurface(self._oneHalfSelected)

    def _deleteConnector(self):
        self._isEditingCubicConnector = False
        items = self._connectorsList.items
        del items[int(self.getSelectedConnectorIndex())]

        cubic = self._cubicConnectors[int(self.getSelectedConnectorIndex())]
        (cubicA, cubicB, _, _, _, _, _) = cubic
        cubicA.setParent(None)
        cubicB.setParent(None)
        cubicA.setVisible(False)
        cubicB.setVisible(False)
        del self._cubicConnectors[int(self.getSelectedConnectorIndex())]

        self._connectorsList.setItems(items)
        self.propertyChanged.emit()


    def _toggleCubicConnectors3DAndPlanar(self, to3D = False):
        if self._areCubicConnectors3D == to3D:
            return
        rootNode = Application.getInstance().getController().getScene().getRoot()
        if to3D:
            self._areCubicConnectors3D = True
            for cubic in self._cubicConnectors:
                (cubicA, cubicB, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = cubic
                depthCA = depthA if isFemaleA else depthB
                depthCB = depthB if isFemaleB else depthA
                cubicA.setScale(Vector(depthCA/0.5 ,cubicA.getScale().y, cubicA.getScale().z))
                cubicB.setScale(Vector(depthCB/0.5 ,cubicB.getScale().y, cubicB.getScale().z))

                fA = 1.0 if isFemaleA else -1.0
                fB = 1.0 if isFemaleB else -1.0

                cubicA.setPosition(Vector(self._cutsurfacePosA.x+(depthCA/2.0)*fA, cubicA.getWorldPosition().y, cubicA.getWorldPosition().z))
                cubicB.setPosition(Vector(self._cutsurfacePosB.x-(depthCB/2.0)*fB, cubicB.getWorldPosition().y, cubicB.getWorldPosition().z))

        else:
            self._areCubicConnectors3D = False
            for cubic in self._cubicConnectors:
                (cubicA, cubicB, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = cubic
                cubicA.setScale(Vector(1.0 ,cubicA.getScale().y, cubicA.getScale().z))
                cubicB.setScale(Vector(1.0 ,cubicB.getScale().y, cubicB.getScale().z))
                cubicA.setPosition(Vector(self._cutsurfacePosA.x, cubicA.getWorldPosition().y, cubicA.getWorldPosition().z))
                cubicB.setPosition(Vector(self._cutsurfacePosB.x, cubicB.getWorldPosition().y, cubicB.getWorldPosition().z))


    def _setVisibilityOfCubicConnectors(self, visible):
        #Logger.log("i", "nothing here to say %s", self._cubicConnectors)
        rootNode = Application.getInstance().getController().getScene().getRoot()
        for i,cubic in enumerate(self._cubicConnectors):
            if i == self.getSelectedConnectorIndex() and self._isEditingCubicConnector:
                continue
            (cubicA, cubicB, _, _, _, _, _) = cubic
            if self._oneHalfSelected:
                n = rootNode if visible else None
                cubicA.setParent(n)
                cubicB.setParent(None)
            else:
                n = rootNode if visible else None
                cubicA.setParent(None)
                cubicB.setParent(n)

    def _removeAllCubicConnectors(self):
        for cubic in self._cubicConnectors:
            (cubicA, cubicB, _, _, _, _, _) = cubic
            cubicA.setParent(None)
            cubicB.setParent(None)
        self._cubicConnectors = []

    def _addCubicConnector(self):
        minimum_t = min(self._handle.getScale().z*2.5, self._handle.getScale().y*2.5)

        if self._tolerance >= minimum_t:
            self._errorMsg = "Tolerance value is too high."
            self.setShowDepthError(True)
            self.propertyChanged.emit()
            return

        if not self._checksCollidingWithMesh:
            self._activeSubView = 4
            self.propertyChanged.emit()

            self._checksCollidingWithMesh = True
            self._isCollidingWithMesh()
            return
        else:
            self._checksCollidingWithMesh = False

        self.setShowDepthError(False)
        self._removeHandles()

        if self._cameraView != CamView.FRONTAL.value:
            self._switchToCutSurface(self._oneHalfSelected)

        yOffsetA = 0
        yOffsetB = 0
        if self._oneHalfSelected:
            yOffsetB = self._otherHalfNodeYOffset - self._currentHandleYOffset
        else:
            yOffsetA = self._oneHalfNodeYOffset - self._currentHandleYOffset

        nodea = PrettyNode.PrettyNode()
        nodea.setName("cubic1")
        nodea.setSelectable(False)
        nodea.setSolidMesh(AutoparterToolHandle.AutoparterToolHandle.getCubicMesh(self._connectorGender == 0 or self._connectorGender == 1))
        nodea.setScale(Vector(1.0, self._handle.getScale().y, self._handle.getScale().z))
        nodea.setPosition(Vector(self._cutsurfacePosA.x ,self._handle.getWorldPosition().y+yOffsetA, self._handle.getWorldPosition().z))


        nodeb = PrettyNode.PrettyNode()
        nodeb.setName("cubic2")
        nodeb.setSelectable(False)
        nodeb.setSolidMesh(AutoparterToolHandle.AutoparterToolHandle.getCubicMesh(self._connectorGender == 0 or self._connectorGender == 2))
        nodeb.setScale(Vector(1.0, self._handle.getScale().y, self._handle.getScale().z))
        nodeb.setPosition(Vector(self._cutsurfacePosB.x ,self._handle.getWorldPosition().y+yOffsetB, self._handle.getWorldPosition().z))

        if self._isEditingCubicConnector:
            (nodea_, nodeb_, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = self._cubicConnectors[self.getSelectedConnectorIndex()]
            nodea_.setParent(None)
            nodeb_.setParent(None)

        if self._oneHalfSelected:
            nodea.setParent(Application.getInstance().getController().getScene().getRoot())
            nodea.setVisible(True)
        else:
            nodeb.setParent(Application.getInstance().getController().getScene().getRoot())
            nodeb.setVisible(True)

        if self._isEditingCubicConnector:
            addtnlProps = {"tolerance": self._tolerance}
            self._cubicConnectors[self.getSelectedConnectorIndex()] = (nodea, nodeb, self._depthA, self._depthB, self._connectorGender == 0 or self._connectorGender == 1, self._connectorGender == 0 or self._connectorGender == 2, addtnlProps)
            self._isEditingCubicConnector = False
            self._activeSubView = 1
            self.propertyChanged.emit()

        else:
            addtnlProps = {"tolerance": self._tolerance}
            self._cubicConnectors.append((nodea, nodeb, self._depthA, self._depthB, self._connectorGender == 0 or self._connectorGender == 1, self._connectorGender == 0 or self._connectorGender == 2, addtnlProps))
            self._addConnectorToList()

        self._toggleCubicConnectors3DAndPlanar(to3D = True)

    def _addConnectorToList(self):
        items = self._connectorsList.items
        items.append({"connectorType": "Cubic connector", "number": "1"})

        self._connectorsList.setItems(items)
        self._activeSubView = 1
        self.propertyChanged.emit()

    def _finallyCreateConnectorsPressed(self):
        """ Called by UI """
        self._createCubicConnectors()

    def _createCubicConnectors(self):
        if len(self._cubicConnectors) > 0:
            #create cubic connectors
            rectanglesData = []

            for cc in self._cubicConnectors:
                (nodea, nodeb, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = cc
                rectData = (Vector(self._sectionalPlaneX, nodea.getWorldPosition().y-self._oneHalfNodeYOffset, nodea.getWorldPosition().z), nodea.getScale().z*2.5, nodea.getScale().y*2.5, depthA, depthB, isFemaleA, isFemaleB, addtnlProps)
                rectanglesData.append(rectData)

            self.setActiveSubView(3)
            self.propertyChanged.emit()

            job = Jobs.CreateCubicConnectorsJob(rectanglesData, self._oneHalfCutSurfaceData, self._otherHalfCutSurfaceData)
            job.finished.connect(self._onCreateCubicConnectorsJobFinished)
            job.start()

    def _onCreateCubicConnectorsJobFinished(self, job):

        if job.getResult() is None:
            return
        self._cancelPressed()

        (meshDataA, meshDataB, separateConnectorsMeshes) = job.getResult()


        originalName = self._oneHalfNode.getName()
        op = GroupedOperation()
        active_build_plate = CuraApplication.getInstance().getMultiBuildPlateModel().activeBuildPlate
        for i in range(0, len(separateConnectorsMeshes)):
            sepCnctrMesh, depth_r, height_r, width_r, rectPosition_r = separateConnectorsMeshes[i]

            connectorNode = CuraSceneNode()
            connectorNode.setName("Connector"+str(i))
            connectorNode.setSelectable(True)
            connectorNode.setCalculateBoundingBox(True)
            connectorNode.setMeshData(sepCnctrMesh)
            connectorNode.calculateBoundingBoxMesh()

            connectorNode.setScale(Vector(depth_r, height_r, width_r))
            connectorNode.setPosition(rectPosition_r, SceneNode.TransformSpace.World)

            connectorNode.addDecorator(BuildPlateDecorator(active_build_plate))
            connectorNode.addDecorator(SliceableObjectDecorator())

            scene = Application.getInstance().getController().getScene()
            rootNode = scene.getRoot()

            op.addOperation(AddSceneNodeOperation(connectorNode, rootNode))
        op.push()


        self._oneHalfNode.setMeshData(meshDataA)
        self._otherHalfNode.setMeshData(meshDataB)


    ################################
    ##
    # Getter and Setter
    ##
    ################################


    def getCanAddCubicConnector(self):
        return len(self._cubicConnectors) == len(self._connectorsList.items)
    def getCanAddShapedConnector(self):
        return len(self._cubicConnectors) == 0
    def getCanCreateConnectors(self):
        return len(self._cubicConnectors) > 0

    def getIsEditingCubicConnector(self):
        return self._isEditingCubicConnector

    def getCanFinallyAddCubicConnector(self):
        return self._rectHandleWithinPoly and not self._rectHandleCollidesWithOtherCubicConnectors

    def getActiveView(self):
        return self._activeView
    def setActiveView(self, view):
        self._activeView = view

    def getAidActivated(self):
        return self._planeStyle
    def setAidActivated(self, style):
        self._planeStyle = style

    def getIsSlicing(self):
        return self._isSlicing

    def setIsSlicing(self, isSlicing):
        self._isSlicing = isSlicing

    def getActiveSubView(self):
        return self._activeSubView
    def setActiveSubView(self, view):
        self._activeSubView = view

    def getShowError(self):
        return self._showError

    def setShowError(self, showError):
        self._showError = showError

    def getShowDepthError(self):
        return self._showDepthError

    def setShowDepthError(self, y):
        self._showDepthError = y

    def getErrorMessage(self):
        return self._errorMsg

    def setErrorMessage(self, errorMsg):
        self._errorMsg = errorMsg

    def getOneHalfSelected(self):
        return self._oneHalfSelected
    def getCameraView(self):
        return self._cameraView



    def setCameraView(self, vv):
        self._cameraView = int(vv)

        if vv == CamView.FRONTAL.value:
            self._switchToCutSurface(self._oneHalfSelected)
        elif vv == CamView.TOP.value:
            self._switchView()
        elif vv == CamView.BOTTOM.value:
            self._switchView()
        elif vv == CamView.LEFT.value:
            self._switchView()
        elif vv == CamView.RIGHT.value:
            self._switchView()

        self.propertyChanged.emit()


    @staticmethod
    def _parseFloat(str_value: str) -> float:
        try:
            parsed_value = float(str_value)
        except ValueError:
            parsed_value = float(0)
        return parsed_value

    def getTolerance(self):
        return self._tolerance

    def setTolerance(self, tolerance):
        parsed_tol = self._parseFloat(tolerance)
        if parsed_tol <= 0:
            parsed_tol = 0
        self._tolerance = parsed_tol

    def getDepthA(self):
        return self._depthA
    def setDepthA(self, depth):
        parsed_depth = self._parseFloat(depth)
        if parsed_depth <= 0:
            parsed_depth = 0.00001
        self._depthA = parsed_depth

        if(not self._oneHalfSelected):
            self._oneHalfSelected = True
            self._focusOnNode(self._oneHalfNode)

            self._cameraView = CamView.TOP.value
            self._switchView()

            self._setHandleDepth(parsed_depth)
            self._moveHandles(False)
        else:
            #self._oneHalfSelected = True
            self._setHandleDepth(parsed_depth)
            self._moveHandles(False)

    def getDepthB(self):
        return self._depthB

    def setDepthB(self, depth):
        parsed_depth = self._parseFloat(depth)
        if parsed_depth <= 0:
            parsed_depth = 0.00001
        self._depthB = parsed_depth
        if (self._oneHalfSelected):
            self._oneHalfSelected = False
            self._focusOnNode(self._otherHalfNode)

            self._cameraView = CamView.TOP.value
            self._switchView()

            self._setHandleDepth(parsed_depth)
            self._moveHandles(False)
        else:
            self._oneHalfSelected = False
            self._setHandleDepth(parsed_depth)
            self._moveHandles(False)

    def getConnectorGender(self):
        return self._connectorGender

    def setConnectorGender(self, gender):
        self._connectorGender = gender
        self.propertyChanged.emit()

        if self._cameraView == CamView.FRONTAL.value:
            self._setHandleDepth(0.01)
        else:
            self._setHandleDepth(self._getCurrentHandleDepth())

        self._moveHandles(False)

    def getConnectorsList(self):
        return self._connectorsList.items

    def setConnectorsList(self, model):
        self._connectorsList = model

    def getSelectedConnectorIndex(self):
        return int(self._selectedConnectorIndex)

    def setSelectedConnectorIndex(self, index):
        self._selectedConnectorIndex = index
