from UM.Job import Job
from . import trimesh

from UM.Mesh.MeshData import MeshData, calculateNormalsFromIndexedVertices

import math
import time
import numpy as np
import os.path
from UM.Logger import Logger
from UM.Math.Color import Color
from UM.Math.Vector import Vector

from . import Operations

class CalculateAABBTreeJob(Job):

    def __init__(self, mesh) -> None:
        super().__init__()
        self._mesh = mesh

    def run(self) -> None:
        self.setResult(Operations.AABBTree(mesh = self._mesh))

class IsCollidingWithMeshJob(Job):

    def __init__(self, meshDataA, meshDataB, handlePosition, handleScale, depthA, depthB, hasToCheckA, hasToCheckB) -> None:
        super().__init__()
        self._meshDataA = meshDataA
        self._meshDataB = meshDataB
        self._handlePosition = handlePosition
        self._handleScale = handleScale
        self._depthA = depthA
        self._depthB = depthB
        self._hasToCheckA = hasToCheckA
        self._hasToCheckB = hasToCheckB

    def run(self) -> None:

        #from trimesh.collision import CollisionManager

        """
            tests if the mesh collides with the cubic connector

            we can't use the CollisionManager because it needs dependencies :(

            approach:
                we test whether line strings intersect with the mesh or not
                    ls origins: 4 corners and the middle of the cube
                    ls directions: into the mesh
                -there are some cases where this method leads to a wrong result (false negative) but in many cases it is correct (depends on the size of the cc and the mesh)
                -the naive method to check every triangle of the cube and the triangles of the mesh is too slow
                -a better way is to create an aabb tree but we have kept it simple

        """

        from UM.Mesh.MeshBuilder import MeshBuilder
        from shapely.geometry import LineString
        from shapely.geometry import Polygon

        if self._meshDataA is None or self._meshDataB is None:
            return
        (openedMeshA, _, _) = self._meshDataA  # trimesh mesh, trimesh mesh, trimesh path
        (openedMeshB, _, _) = self._meshDataB

        rectPosition = self._handlePosition
        width = self._handleScale.z*2.5
        height = self._handleScale.y*2.5


        rectCornerVertices = [
            [rectPosition.x, rectPosition.y+height/2.0, rectPosition.z-width/2.0], #top left
            [rectPosition.x, rectPosition.y+height/2.0, rectPosition.z+width/2.0], #top right
            [rectPosition.x, rectPosition.y-height/2.0, rectPosition.z+width/2.0], #bottom right
            [rectPosition.x, rectPosition.y-height/2.0, rectPosition.z-width/2.0], #bottom left
            [rectPosition.x, rectPosition.y, rectPosition.z], #center
        ]
        rectCornerVerticesDepthLambda = lambda depth : [
            [rectPosition.x+depth, rectPosition.y+height/2.0, rectPosition.z-width/2.0], #top left
            [rectPosition.x+depth, rectPosition.y+height/2.0, rectPosition.z+width/2.0], #top right
            [rectPosition.x+depth, rectPosition.y-height/2.0, rectPosition.z+width/2.0], #bottom right
            [rectPosition.x+depth, rectPosition.y-height/2.0, rectPosition.z-width/2.0], #bottom left
            [rectPosition.x+depth, rectPosition.y, rectPosition.z], #center
        ]


        intersects = [False, False]

        for z in range(0,2):
            if not self._hasToCheckA and z == 0:
                continue
            if not self._hasToCheckB and z == 1:
                continue

            if z == 0:
                rectCornerVerticesDepth = rectCornerVerticesDepthLambda(self._depthA)
                openedMesh = openedMeshA
            else:
                rectCornerVerticesDepth = rectCornerVerticesDepthLambda(-self._depthB)
                openedMesh = openedMeshB

            lines = []
            for i in range(0, len(rectCornerVertices)):
                line = LineString([(rectCornerVertices[i][0], rectCornerVertices[i][1], rectCornerVertices[i][2]), (rectCornerVerticesDepth[i][0], rectCornerVerticesDepth[i][1], rectCornerVerticesDepth[i][2])])
                lines.append(line)

            vertices = openedMesh.vertices


            for i in range(0, len(lines)):
                line = lines[i]
                _, lineY, lineZ = line.coords[0]
                possible_indices = []
                #find all triangles which could intersect with the line (filter y,z)


                for j in range(0, len(openedMesh.faces)):
                    elimination = [0,0,0,0]
                    for k in range(0, 3):
                        vv = vertices[openedMesh.faces[j][k]]
                        if vv[1] > lineY:
                            elimination[0] += 1
                        if vv[1] < lineY:
                            elimination[1] += 1
                        if vv[2] > lineZ:
                            elimination[2] += 1
                        if vv[2] < lineZ:
                            elimination[3]+= 1
                    not_possible = False
                    for k in range(0, 4):
                        if elimination[k] >= 3:
                            not_possible = True
                    if not not_possible:
                        possible_indices.append(j)
                #test all triangles which could intersect
                for j in range(0, len(possible_indices)):

                    tri = [vertices[openedMesh.faces[possible_indices[j]][0]], vertices[openedMesh.faces[possible_indices[j]][1]], vertices[openedMesh.faces[possible_indices[j]][2]]]
                    tri_poly = Polygon([[tri[0][0], tri[0][1], tri[0][2]], [tri[1][0], tri[1][1], tri[1][2]], [tri[2][0], tri[2][1], tri[2][2]], [tri[0][0], tri[0][1], tri[0][2]]])

                    if(tri_poly.intersects(line)):
                        intersects[z] = True

        #if intersects[0] or intersects[1]:
        #    Logger.log("i", "intersects YES")
        #else:
        #    Logger.log("i", "it does not intersect :))")



        self.setResult((intersects[0], intersects[1]))

class IsCollidingWithOtherRectsJob(Job):

    def __init__(self, handle, cubicConnectors, oneHalfSelected) -> None:
        super().__init__()
        self._handle = handle
        self._cubicConnectors = cubicConnectors
        self._oneHalfSelected = oneHalfSelected

    def run(self) -> None:

        if self._handle is None:
            return

        from shapely.geometry import Polygon

        rectPosition = self._handle.getWorldPosition()
        width = self._handle.getScale().z*2.5
        height = self._handle.getScale().y*2.5

        rectCornerVertices = [
            [rectPosition.z-width/2.0, rectPosition.y+height/2.0], #top left
            [rectPosition.z+width/2.0, rectPosition.y+height/2.0], #top right
            [rectPosition.z+width/2.0, rectPosition.y-height/2.0], #bottom right
            [rectPosition.z-width/2.0, rectPosition.y-height/2.0], #bottom left
        ]
        handleRectanglePolygon =  Polygon([[rectCornerVertices[0][0], rectCornerVertices[0][1]], [rectCornerVertices[1][0], rectCornerVertices[1][1]], [rectCornerVertices[2][0], rectCornerVertices[2][1]], [rectCornerVertices[3][0], rectCornerVertices[3][1]], [rectCornerVertices[0][0], rectCornerVertices[0][1]]])

        collide = False

        for cubic in self._cubicConnectors:
            (cubicA, cubicB, _, _, _, _, _) = cubic
            cubicNode = cubicA if self._oneHalfSelected else cubicB
            rectPosition = cubicNode.getWorldPosition()
            width = cubicNode.getScale().z*2.5
            height = cubicNode.getScale().y*2.5
            rectCornerVertices = [
                [rectPosition.z-width/2.0, rectPosition.y+height/2.0], #top left
                [rectPosition.z+width/2.0, rectPosition.y+height/2.0], #top right
                [rectPosition.z+width/2.0, rectPosition.y-height/2.0], #bottom right
                [rectPosition.z-width/2.0, rectPosition.y-height/2.0], #bottom left
            ]
            rectanglePolygon =  Polygon([[rectCornerVertices[0][0], rectCornerVertices[0][1]], [rectCornerVertices[1][0], rectCornerVertices[1][1]], [rectCornerVertices[2][0], rectCornerVertices[2][1]], [rectCornerVertices[3][0], rectCornerVertices[3][1]], [rectCornerVertices[0][0], rectCornerVertices[0][1]]])

            if not rectanglePolygon.disjoint(handleRectanglePolygon):
                collide = True

        self.setResult(collide)

class IsRectInPolyJob(Job):

    def __init__(self, rect, meshData) -> None:
        super().__init__()
        self._rect = rect
        self._mesh = meshData

    def run(self) -> None:

        if self._mesh is None:
            return

        (_, _, onPlaneAnd3D) = self._mesh  # trimesh mesh, trimesh mesh, trimesh path
        (rectPosition, width, height) = self._rect

        #width is depth -> z
        #height is height ->y
        #Logger.log("i", "ich bin ein job %s", points2d)

        plane_origin = [rectPosition.x,0,0]
        plane_normal = [1.0,0,0]

        allCornerVertices = [
            [rectPosition.x, rectPosition.y+height/2.0, rectPosition.z-width/2.0], #top left
            [rectPosition.x, rectPosition.y+height/2.0, rectPosition.z+width/2.0], #top right
            [rectPosition.x, rectPosition.y-height/2.0, rectPosition.z-width/2.0], #bottom left
            [rectPosition.x, rectPosition.y-height/2.0, rectPosition.z+width/2.0] #bottom right
        ]

        isInPoly = Operations.is_vertex_in_poly(onPlaneAnd3D, allCornerVertices, (plane_origin, plane_normal))

        self.setResult(isInPoly)

class CreateCubicConnectorsJob(Job):

    def __init__(self, rectanglesData, meshDataA, meshDataB) -> None:
        super().__init__()
        self._rectanglesData = rectanglesData
        self._meshDataA = meshDataA
        self._meshDataB = meshDataB

    def _createCube(self, size):
        from UM.Mesh.MeshBuilder import MeshBuilder
        mesh = MeshBuilder()

        # Can't use MeshBuilder.addCube() because that does not get per-vertex normals
        # Per-vertex normals require duplication of vertices
        s = size / 2
        verts = [ # 6 faces with 4 corners each
            [-s, -s,  s], [-s,  s,  s], [ s,  s,  s], [ s, -s,  s],
            [-s,  s, -s], [-s, -s, -s], [ s, -s, -s], [ s,  s, -s],
            [ s, -s, -s], [-s, -s, -s], [-s, -s,  s], [ s, -s,  s],
            [-s,  s, -s], [ s,  s, -s], [ s,  s,  s], [-s,  s,  s],
            [-s, -s,  s], [-s, -s, -s], [-s,  s, -s], [-s,  s,  s],
            [ s, -s, -s], [ s, -s,  s], [ s,  s,  s], [ s,  s, -s]
        ]
        mesh.setVertices(np.asarray(verts, dtype=np.float32))

        indices = []
        for i in range(0, 24, 4): # All 6 quads (12 triangles)
            indices.append([i, i+2, i+1])
            indices.append([i, i+3, i+2])
        mesh.setIndices(np.asarray(indices, dtype=np.int32))

        mesh.calculateNormals()
        return mesh.build()


    def run(self) -> None:

        if self._meshDataA is None or self._meshDataB is None:
            return

        (openedMeshA, _, cutsurfaceOnPlaneA) = self._meshDataA  # trimesh mesh, trimesh mesh, trimesh path
        (openedMeshB, _, cutsurfaceOnPlaneB) = self._meshDataB

        (rectPosition, width, height, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = self._rectanglesData[0]
        #width is depth -> z
        #height is height ->y
        plane_origin = [rectPosition.x,0,0]
        plane_normal = [1.0,0,0]


        addCubicConnectors = Operations.add_cubic_connectors(cutsurfaceOnPlaneA, self._rectanglesData, (plane_origin, plane_normal), openedMeshA, openedMeshB)


        if addCubicConnectors is not None:

            #in the case that separate connectors exists
            separateConnectorsMeshes = []
            for rect in self._rectanglesData:
                (rectPosition, width_r, height, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = rect
                if isFemaleA and isFemaleB:
                    tolerance = addtnlProps["tolerance"]
                    separateConnectorsMeshes.append((self._createCube(size=1.0), (depthA+depthB), height-tolerance, width_r-tolerance, rectPosition))


            oneHalfTrimesh, otherHalfTrimesh = addCubicConnectors

            oneHalfMesh = Operations.toMeshData(oneHalfTrimesh)
            otherHalfMesh = Operations.toMeshData(otherHalfTrimesh)
            self.setResult((oneHalfMesh, otherHalfMesh, separateConnectorsMeshes))
        else:
            self.setResult(addCubicConnectors)



class SliceMeshJob(Job):


    def __init__(self, mesh: MeshData, normal_vector: [], origin_vector: []) -> None:
        super().__init__()
        self.mesh = mesh
        self.normal_vector = normal_vector
        self.origin_vector = origin_vector

    def run(self) -> None:

        if self.mesh is None:
            return


        # convert to trimesh object
        tmesh = Operations.toTriMesh(self.mesh)

        #slice_mesh_plane_v2 is a self modified version of slice_mesh_plane which doesn't need more dependencies but is a little slower

        oneHalf, oneHalfCutSurfaceData = Operations.slice_mesh(tmesh, [1.0, 0.0, 0.0], [self.origin_vector[0],0,0], cap=True)
        otherHalf, otherHalfCutSurfaceData = Operations.slice_mesh(tmesh, [-1.0, 0.0, 0.0], [self.origin_vector[0],0,0], cap=True)

        oneHalf.fill_holes()

        otherHalf.fill_holes()

        # convert Trimesh to MeshData
        oneHalfMesh = Operations.toMeshData(oneHalf)
        otherHalfMesh = Operations.toMeshData(otherHalf)


        self.setResult(((oneHalfMesh, oneHalfCutSurfaceData), (otherHalfMesh, otherHalfCutSurfaceData)))
