import numpy as np
from . import trimesh
from .trimesh import util
from .trimesh import geometry
from .trimesh import grouping
from .trimesh import transformations

from .trimesh import points
from shapely.geometry import Polygon

from .trimesh.base import Trimesh
from UM.Logger import Logger

from UM.Math.Color import Color
from UM.Math.Vector import Vector
from UM.Mesh.MeshData import MeshData, calculateNormalsFromIndexedVertices

# custom function
def toTriMesh(mesh_data: MeshData) -> trimesh.base.Trimesh:
    vertices = mesh_data.getVertices()
    indices = mesh_data.getIndices()
    if indices is None:
        indices = []
        for face_id in range(0, (int(len(vertices)/3))):
            base_index = face_id * 3
            v_a = base_index#verts[base_index]
            v_b = base_index + 1 #verts[base_index + 1]
            v_c = base_index + 2 #verts[base_index + 2]
            indices.append([v_a, v_b, v_c])

    return trimesh.base.Trimesh(vertices=vertices, faces=indices)

# The toMeshData function is taken from the AMFReader class which was built by fieldOfView.
def toMeshData(tri_node: trimesh.base.Trimesh) -> MeshData:
    tri_faces = tri_node.faces
    tri_vertices = tri_node.vertices

    indices = []
    vertices = []

    index_count = 0
    face_count = 0
    for tri_face in tri_faces:
        face = []
        for tri_index in tri_face:
            vertices.append(tri_vertices[tri_index])
            face.append(index_count)
            index_count += 1
        indices.append(face)
        face_count += 1

    vertices = np.asarray(vertices, dtype = np.float32)
    indices = np.asarray(indices, dtype = np.int32)
    normals = calculateNormalsFromIndexedVertices(vertices, indices, face_count)

    mesh_data = MeshData(vertices = vertices, indices = indices, normals = normals)
    return mesh_data

def colorizeMesh(mesh_data: MeshData, color: Color) -> MeshData:

    colors = np.zeros((10, 4), dtype = np.float32)
    colors.resize((len(mesh_data.getVertices()), 4), refcheck = False)

    for i in range(len(mesh_data.getVertices())):
        colors[i, 0] = color.r/255.0
        colors[i, 1] = color.g/255.0
        colors[i, 2] = color.b/255.0
        colors[i, 3] = color.a

    return mesh_data.set(colors=colors)

"""def isFaceCollidingWithMesh_AABB(aabbTree = None, face = [[0,0,0], [0,0,10], [0,5,10]):

    bigger = lambda a,b : a[0] >= b[0] and a[1] >= b[1] and a[2] >= b[2]
    smaller = lambda a,b : a[0] <= b[0] and a[1] <= b[1] and a[2] <= b[2]
    isWithin = lambda min,max, p : bigger(p, min) and smaller(p, max)

    while not (aabbTree is None):
        (minx, miny, minz), (maxx, maxy, maxz) = aabbTree.bounds
        within = False
        for i in range(0,3):
            if isWithin([minx, miny, minz], [maxx, maxy, maxz], face[i]):
                within = True
        if within:
            #check aabbTree
            #maybe check aabbTree.a
            #maybe check aabbTree.b
        else:"""


class AABBTree(object):

    """
        -must be initiated with a trimesh mesh
        -primitive approach to map the faces
        -recursion depth is limited

        -it's not in use yet
        -used for spatial indexing
        -the method isn't based on a popular approach etc.

        splitDirection = 0 : x
        splitDirection = 1 : y
        splitDirection = 2 : z
    """

    def __init__(self, mesh = None, splitDirection = 0, depth = 15):
        super().__init__()
        self.a = None
        self.b = None
        self.bounds = mesh.bounds
        self.mesh = None
        depth -= 1


        if mesh is None:
            return

        if len(mesh.faces) < 5 or depth < 1:
            self.mesh = mesh
            return

        a_1,b_1,m_1 = self.split(mesh, 0)
        a_2,b_2,m_2 = self.split(mesh, 1)
        a_3,b_3,m_3 = self.split(mesh, 2)


        #we want to split as many faces as possible into different bbs

        if len(m_1.faces) <= len(m_2.faces) and len(m_1.faces) <= len(m_3.faces):
            self.a = AABBTree(mesh = a_1, splitDirection = 0, depth=depth) if len(a_1.faces) > 0 else None
            self.b = AABBTree(mesh = b_1,  splitDirection = 0, depth=depth) if len(b_1.faces) > 0 else None
            self.mesh = m_1
        elif len(m_2.faces) <= len(m_1.faces) and len(m_2.faces) <= len(m_3.faces):

            self.a = AABBTree(mesh = a_2, splitDirection = 1, depth=depth) if len(a_2.faces) > 0 else None
            self.b = AABBTree(mesh = b_2,  splitDirection = 1, depth=depth) if len(b_2.faces) > 0 else None
            self.mesh = m_2
        else:

            if len(m_1.faces) == len(mesh.faces) and len(m_2.faces) == len(mesh.faces) and len(m_3.faces) == len(mesh.faces):
                self.mesh = mesh
                return

            self.a = AABBTree(mesh = a_3, splitDirection = 2, depth=depth) if len(a_3.faces) > 0 else None
            self.b = AABBTree(mesh = b_3,  splitDirection = 2, depth=depth) if len(b_3.faces) > 0 else None
            self.mesh = m_3


    def split(self, mesh, splitDirection):

        (minx, miny, minz), (maxx, maxy, maxz) = mesh.bounds

        a_min, a_max = [minx, miny, minz], []
        if splitDirection == 0:
            a_max = [minx+(maxx-minx)/2, maxy, maxz]
        elif splitDirection == 1:
            a_max = [maxx, miny+(maxy-miny)/2, maxz]
        else:
            a_max = [maxx, maxy, minz+(maxz-minz)/2]

        b_min, b_max = [], [maxx, maxy, maxz]

        if splitDirection == 0:
            b_min = [minx+(maxx-minx)/2, miny, minz]
        elif splitDirection == 1:
            b_min = [minx, miny+(maxy-miny)/2, minz]
        else:
            b_min = [minx, miny, minz+(maxz-minz)/2]


        a_faces_indices = []
        b_faces_indices = []
        m_faces_indices = []

        for j in range(0, len(mesh.faces)):
            f = mesh.faces[j]
            within_a = True
            within_b = True
            for i in f:
                v = mesh.vertices[i]
                if not ((v[0] >= a_min[0] and v[1] >= a_min[1] and v[2] >= a_min[2]) and (v[0] <= a_max[0] and v[1] <= a_max[1] and v[2] <= a_max[2])):
                    #not in a
                    within_a = False
                elif not ((v[0] >= b_min[0] and v[1] >= b_min[1] and v[2] >= b_min[2]) and (v[0] <= b_max[0] and v[1] <= b_max[1] and v[2] <= b_max[2])):
                    #not in b
                    within_b = False

            if within_a:
                a_faces_indices.append(j)
            elif within_b:
                b_faces_indices.append(j)
            else:
                m_faces_indices.append(j)
        # we know now all faces
        a_faces,b_faces,m_faces = [], [], []
        for idx in a_faces_indices:
            a_faces.append(mesh.faces[idx])
        for idx in b_faces_indices:
            b_faces.append(mesh.faces[idx])
        for idx in m_faces_indices:
            m_faces.append(mesh.faces[idx])



        return Trimesh(vertices=mesh.vertices, faces=a_faces), Trimesh(vertices=mesh.vertices, faces=b_faces), Trimesh(vertices=mesh.vertices, faces=m_faces)



def slice_mesh(mesh,
                     plane_normal,
                     plane_origin,
                     cap=False,
                     cached_dots=None,
                     **kwargs):
    """

    !!! this function has been taken from trimesh.intersections and modified

    Slice a mesh with a plane, returning a new mesh that is the
    portion of the original mesh to the positive normal side of the plane

    Parameters
    ---------
    mesh : Trimesh object
      Source mesh to slice
    plane_normal : (3,) float
      Normal vector of plane to intersect with mesh
    plane_origin :  (3,) float
      Point on plane to intersect with mesh
    cap : bool
      If True, cap the result with a triangulated polygon
    cached_dots : (n, 3) float
        If an external function has stored dot
        products pass them here to avoid recomputing

    Returns
    ----------
    new_mesh : Trimesh object
      Sliced mesh
    new_mesh_opened : Trimesh object
      Sliced mesh without cap
    cutsurface_mesh : Trimesh object
      the cap as a mesh
    onPlaneAnd3D : Path
      the section as polygons on the xy plane

    """
    # check input for none
    if mesh is None:
        return None

    # avoid circular import
    from .trimesh.base import Trimesh
    from .trimesh.creation import triangulate_polygon

    # check input plane
    plane_normal = np.asanyarray(plane_normal,
                                 dtype=np.float64)
    plane_origin = np.asanyarray(plane_origin,
                                 dtype=np.float64)

    # check to make sure origins and normals have acceptable shape
    shape_ok = ((plane_origin.shape == (3,) or
                 util.is_shape(plane_origin, (-1, 3))) and
                (plane_normal.shape == (3,) or
                 util.is_shape(plane_normal, (-1, 3))) and
                plane_origin.shape == plane_normal.shape)
    if not shape_ok:
        raise ValueError('plane origins and normals must be (n, 3)!')

    # start with copy of original mesh, faces, and vertices
    sliced_mesh = mesh.copy()
    vertices = mesh.vertices.copy()
    faces = mesh.faces.copy()

    # slice away specified planes
    for origin, normal in zip(plane_origin.reshape((-1, 3)),
                              plane_normal.reshape((-1, 3))):

        # calculate dots here if not passed in to save time
        # in case of cap
        if cached_dots is None:
            # dot product of each vertex with the plane normal indexed by face
            # so for each face the dot product of each vertex is a row
            # shape is the same as faces (n,3)
            dots = np.einsum('i,ij->j', normal,
                             (vertices - origin).T)[faces]
        else:
            dots = cached_dots
        # save the new vertices and faces
        vertices, faces = trimesh.intersections.slice_faces_plane(vertices=vertices,
                                            faces=faces,
                                            plane_normal=normal,
                                            plane_origin=origin,
                                            cached_dots=dots)


        if cap:
            # check if mesh is watertight (can't cap if not)
            #if not sliced_mesh.is_watertight:
            #    raise ValueError('Input mesh must be watertight to cap slice')

            path = sliced_mesh.section(plane_normal=normal,
                                       plane_origin=origin,
                                       cached_dots=dots)


            #cutsurface plane

            to_2D = geometry.plane_transform(origin=origin,normal=normal)

            # transform Path3D onto XY plane for triangulation

            on_plane, to_3D = path.to_planar(to_2D=to_2D)

            v, f = [], []

            facecount = 0

            from shapely.ops import triangulate

            for polygon in on_plane.polygons_closed:

                triangles = triangulate(polygon)

                for triangle in triangles:

                    if not triangle.within(polygon):
                        continue
                    x, y = triangle.exterior.coords.xy

                    for i in range(0, 3):
                        v.append([x[i], y[i]])
                    f.append([facecount*3+2, facecount*3+1, facecount*3])
                    facecount = facecount + 1


            vf, ff = util.append_faces(v, f)
            vf = np.column_stack((vf, np.zeros(len(vf))))
            vf = transformations.transform_points(vf, to_3D)

            #save opened meshes so that we can use them later to add connectors

            openedMesh = Trimesh(vertices=vertices.copy(), faces=faces.copy(), process=False)

            #save cut surface as mesh for later use

            cutSurfaceMesh = Trimesh(vertices=vf, faces=f, process=False)

            #append cap to the sliced mesh

            vertices, faces = util.append_faces([vertices, vf], [faces, f])

            #processing (combine vertices and reindex / takes some time) (makes it also watertight)

            sliced_mesh = Trimesh(vertices=vertices, faces=faces)
            vertices, faces = sliced_mesh.vertices.copy(), sliced_mesh.faces.copy()

            onPlaneAnd3D = (on_plane, to_3D)


    return (Trimesh(vertices=vertices, faces=faces, process=False), (openedMesh, cutSurfaceMesh, onPlaneAnd3D))




def is_vertex_in_poly(polygon_on_xyplane, vertices, plane):

    """ checks if given vertices are within polygon """

    from UM.Logger import Logger
    from shapely.geometry import Point

    (plane_origin, plane_normal) = plane

    (points2d, _) = points.project_to_plane(vertices, plane_normal=plane_normal, plane_origin=plane_origin, return_transform=True)

    (on_plane, _) = polygon_on_xyplane

    for polygon in on_plane.polygons_closed:
        contains = True
        for p in points2d:
            result = not polygon.disjoint(Point(p[0], p[1]))
            if not result:
                contains = False
                break
        if contains:
            return True

    return False



def cut_rectangle_out_from_polygons(rectangle = None, polygons = []):

    """
    cuts a given rectangle out of given polygons and returns the newly emerged polygons
    (2D)

    Parameters:
        rectangle : [[leftTop.x, leftTop.y], [rightTop.x, rightTop.y], [rightBottom.x, rightBottom.y], [leftBottom.x, leftBottom.y]]
        polygons : Path 2D
    Returns:
        a set of polygons : [Polygon]
    """

    if rectangle is None:
        return polygons

    topLeftCorner, topRightCorner, bottomRightCorner, bottomLeftCorner = rectangle

    rectanglePolygon =  Polygon([[topLeftCorner[0], topLeftCorner[1]], [topRightCorner[0], topRightCorner[1]], [bottomRightCorner[0], bottomRightCorner[1]], [bottomLeftCorner[0], bottomLeftCorner[1]], [topLeftCorner[0], topLeftCorner[1]]])

    x_min, x_max, y_min, y_max = 0,0,0,0
    for i in range(0, len(polygons)):
        x, y = polygons[i].exterior.coords.xy

        if i == 0:
            x_min, x_max, y_min, y_max = min(x)-10, max(x)+10, min(y)-10, max(y)+10
        else:
            x_min, x_max, y_min, y_max = min(x_min, min(x)-10), max(x_max, max(x)+10), min(y_min, min(y)-10), max(y_max, max(y)+10)

    from shapely.ops import split
    from shapely.geometry import LineString
    from shapely.ops import triangulate

    res = []
    for pl in polygons:
        res +=  list((split(pl, LineString([(x_min, topLeftCorner[1]), (x_max, topLeftCorner[1])]))).geoms)
    polygons = res

    res = []
    for pl in polygons:
        res +=  list((split(pl, LineString([(x_min, bottomLeftCorner[1]), (x_max, bottomLeftCorner[1])]))).geoms)
    polygons = res

    res = []
    for pl in polygons:
        res +=  list((split(pl, LineString([(topLeftCorner[0], y_min), (topLeftCorner[0], y_max)]))).geoms)
    polygons = res

    res = []
    for pl in polygons:
        res +=  list((split(pl, LineString([(topRightCorner[0], y_min), (topRightCorner[0], y_max)]))).geoms)
    polygons = res

    #get polygons which are the rectangle or a piece of it
    polygonsInRect = []
    for i in range(0, len(polygons)):
        pl = polygons[i]
        if pl.almost_equals(rectanglePolygon) or rectanglePolygon.contains(pl):
            polygonsInRect.append(i)
    #delete those polygons
    for index in sorted(polygonsInRect, reverse=True):
        del polygons[index]

    return polygons



def calculate_cubic_cap(rectData = None, forPartA = True):

    """ calculates a cubic cap """

    from UM.Logger import Logger

    if rectData is None:
        return None
    (rectPosition, width, height, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = rectData

    if (forPartA and not isFemaleA) or (not forPartA and not isFemaleB):
        width, height = subtract_tolerance(width, height, addtnlProps)

    rectCornerVertices = [
        [rectPosition.x, rectPosition.y+height/2.0, rectPosition.z-width/2.0], #top left
        [rectPosition.x, rectPosition.y+height/2.0, rectPosition.z+width/2.0], #top right
        [rectPosition.x, rectPosition.y-height/2.0, rectPosition.z-width/2.0], #bottom left
        [rectPosition.x, rectPosition.y-height/2.0, rectPosition.z+width/2.0] #bottom right
    ]
    topLeftCorner, topRightCorner, bottomLeftCorner, bottomRightCorner  = rectCornerVertices[0], rectCornerVertices[1], rectCornerVertices[2], rectCornerVertices[3]

    if forPartA:
        extrudeDirection = 1 if isFemaleA else -1
        depth = depthA if isFemaleA else depthB

        depth *= extrudeDirection
        rectDepthCornerVertices = [
            [rectCornerVertices[0][0]+depth, rectCornerVertices[0][1], rectCornerVertices[0][2]], #top left
            [rectCornerVertices[1][0]+depth, rectCornerVertices[1][1], rectCornerVertices[1][2]], #top right
            [rectCornerVertices[2][0]+depth, rectCornerVertices[2][1], rectCornerVertices[2][2]], #bottom left
            [rectCornerVertices[3][0]+depth, rectCornerVertices[3][1], rectCornerVertices[3][2]] #bottom right
        ]
        #[::rvrs] reverses the array - a face is only visible if the vertices are in a clockwise order
        rvrs = -1
        topLeftCornerDepth, topRightCornerDepth, bottomLeftCornerDepth, bottomRightCornerDepth  = rectDepthCornerVertices[0], rectDepthCornerVertices[1], rectDepthCornerVertices[2], rectDepthCornerVertices[3]
        leftWall = [[topLeftCorner, bottomLeftCornerDepth, bottomLeftCorner][::rvrs], [topLeftCornerDepth, bottomLeftCornerDepth,topLeftCorner][::rvrs]]
        upperWall = [[topRightCorner, topLeftCornerDepth, topLeftCorner][::rvrs], [topRightCornerDepth, topLeftCornerDepth, topRightCorner][::rvrs]]
        rightWall = [[topRightCornerDepth, topRightCorner, bottomRightCorner][::rvrs], [bottomRightCornerDepth, topRightCornerDepth, bottomRightCorner][::rvrs]]
        lowerWall = [[bottomLeftCorner, bottomLeftCornerDepth, bottomRightCorner][::rvrs], [bottomRightCorner, bottomLeftCornerDepth, bottomRightCornerDepth][::rvrs]]

        frontalWall = [[bottomLeftCornerDepth, topLeftCornerDepth, topRightCornerDepth][::rvrs], [bottomLeftCornerDepth, topRightCornerDepth, bottomRightCornerDepth][::rvrs]]

        return frontalWall+upperWall+lowerWall+leftWall+rightWall
    else:
        extrudeDirection = -1 if isFemaleB else 1
        depth = depthB if isFemaleB else depthA
        depth *= extrudeDirection
        rectDepthCornerVertices = [
            [rectCornerVertices[0][0]+depth, rectCornerVertices[0][1], rectCornerVertices[0][2]], #top left
            [rectCornerVertices[1][0]+depth, rectCornerVertices[1][1], rectCornerVertices[1][2]], #top right
            [rectCornerVertices[2][0]+depth, rectCornerVertices[2][1], rectCornerVertices[2][2]], #bottom left
            [rectCornerVertices[3][0]+depth, rectCornerVertices[3][1], rectCornerVertices[3][2]] #bottom right
        ]
        rvrs = 1
        topLeftCornerDepth, topRightCornerDepth, bottomLeftCornerDepth, bottomRightCornerDepth  = rectDepthCornerVertices[0], rectDepthCornerVertices[1], rectDepthCornerVertices[2], rectDepthCornerVertices[3]
        leftWall = [[topLeftCorner, bottomLeftCornerDepth, bottomLeftCorner][::rvrs], [topLeftCornerDepth, bottomLeftCornerDepth,topLeftCorner][::rvrs]]
        upperWall = [[topRightCorner, topLeftCornerDepth, topLeftCorner][::rvrs], [topRightCornerDepth, topLeftCornerDepth, topRightCorner][::rvrs]]
        rightWall = [[topRightCornerDepth, topRightCorner, bottomRightCorner][::rvrs], [bottomRightCornerDepth, topRightCornerDepth, bottomRightCorner][::rvrs]]
        lowerWall = [[bottomLeftCorner, bottomLeftCornerDepth, bottomRightCorner][::rvrs], [bottomRightCorner, bottomLeftCornerDepth, bottomRightCornerDepth][::rvrs]]

        frontalWall = [[bottomLeftCornerDepth, topLeftCornerDepth, topRightCornerDepth][::rvrs], [bottomLeftCornerDepth, topRightCornerDepth, bottomRightCornerDepth][::rvrs]]

        return frontalWall+upperWall+lowerWall+leftWall+rightWall


def subtract_tolerance(width, height, addtnlProps):
    tolerance = addtnlProps["tolerance"]
    return width-tolerance, height-tolerance

def add_cubic_connectors(cutsurfaceOnPlaneA, rectanglesData, rectanglesPlane, openedMeshA, openedMeshB, **kwargs):
    from .trimesh.base import Trimesh
    from .trimesh.creation import triangulate_polygon

    from UM.Logger import Logger

    (plane_origin, plane_normal) = rectanglesPlane

    rectanglesCornerVertices2D_A = []
    rectanglesCornerVertices2D_B = []

    for a_b in range(0, 2):
        #we need to do this separately because of the tolerance
        #a_b = 0: for part a
        #a_b = 1: for part b
        for rectData in rectanglesData:
            (rectPosition, width, height, depthA, depthB, isFemaleA, isFemaleB, addtnlProps) = rectData
            if (a_b == 0 and not isFemaleA) or (a_b == 1 and not isFemaleB):
                width, height = subtract_tolerance(width, height, addtnlProps)

            rectCornerVertices = [
                [rectPosition.x, rectPosition.y+height/2.0, rectPosition.z-width/2.0], #top left
                [rectPosition.x, rectPosition.y+height/2.0, rectPosition.z+width/2.0], #top right
                [rectPosition.x, rectPosition.y-height/2.0, rectPosition.z-width/2.0], #bottom left
                [rectPosition.x, rectPosition.y-height/2.0, rectPosition.z+width/2.0] #bottom right
            ]
            #project corner vertices of rect on xy plane
            (points2d, _) = points.project_to_plane(rectCornerVertices, plane_normal=plane_normal, plane_origin=plane_origin, return_transform=True)
            if a_b == 0:
                rectanglesCornerVertices2D_A.append(points2d)
            else:
                rectanglesCornerVertices2D_B.append(points2d)


    #cut surface polygons 2d
    (cutsurface_2d_points, cutsurface_to_3D) = cutsurfaceOnPlaneA



    from shapely.geometry import Point
    # iterate through all polygons which represent the cut surface

    f_a, v_a =  [], []
    f_b, v_b = [], []


    for a_b in range(0, 2):
        cutted_polygons = []

        #we need to do this separately because of the tolerance
        if a_b == 0:
            rectanglesCornerVertices2D = rectanglesCornerVertices2D_A
        else:
            rectanglesCornerVertices2D = rectanglesCornerVertices2D_B

        for closed_polygon in cutsurface_2d_points.polygons_closed:
            contained_rects_indices = []
            # find cubic connectors inside of this polygon and save indices
            for i in range(0, len(rectanglesCornerVertices2D)):

                contains = True
                for p in rectanglesCornerVertices2D[i]:
                    result = not closed_polygon.disjoint(Point(p[0], p[1]))
                    if not result:
                        contains = False
                        break
                if contains:
                    contained_rects_indices.append(i)

            #now we know every cubic connector which is inside of this polygon
            #cut them out of this polygon


            polys = [closed_polygon]
            for j in range(0, len(contained_rects_indices)):
                points2d = rectanglesCornerVertices2D[contained_rects_indices[j]]
                topLeftCorner = points2d[0]
                topRightCorner = points2d[1]
                bottomLeftCorner = points2d[2]
                bottomRightCorner = points2d[3]

                polys = cut_rectangle_out_from_polygons(rectangle = (topLeftCorner, topRightCorner, bottomRightCorner, bottomLeftCorner), polygons = polys)
            #add the resulting polygons into the set of all final polygons for further triangulation

            cutted_polygons += polys


        from shapely.ops import triangulate

        v, f = [], []
        f_rev = []
        facecount = 0
        #iterate through all cutted polygons, triangulate them and create the 2d cutsurface mesh
        for i in range(0, len(cutted_polygons)):

            current_poly = cutted_polygons[i]
            triangles = triangulate(current_poly)

            for triangle in triangles:

                if not triangle.within(current_poly):
                    continue
                x, y = triangle.exterior.coords.xy
                for i in range(0, 3):
                    v.append([x[i], y[i]])
                #reversed
                f.append([facecount*3+2, facecount*3+1, facecount*3])
                #normal (but ironically we call them 'reversed' ;))
                f_rev.append([facecount*3, facecount*3+1, facecount*3+2])
                facecount = facecount + 1


        vf, ff = util.append_faces(v, f)
        vf = np.column_stack((vf, np.zeros(len(vf))))
        # transform points back from xy plane to yz plane
        vf = transformations.transform_points(vf, cutsurface_to_3D)
        if a_b == 0:
            v_a = vf
            f_a = f
        else:
            v_b = vf
            f_b = f_rev

    # our mesh has the desired holes - we need to cap them with the right depths

    cubicCapsA = []
    cubicCapsB = []
    for rectData in rectanglesData:
        cubicCapsA += calculate_cubic_cap(rectData=rectData, forPartA=True)
        cubicCapsB += calculate_cubic_cap(rectData=rectData, forPartA=False)

    # add caps to the mesh

    #create cap-mesh-a

    v_capA, f_capA = [], []
    facecount = 0
    for capTriangle in cubicCapsA:
        for i in range(0, 3):
            v_capA.append([capTriangle[i][0], capTriangle[i][1], capTriangle[i][2]])
        f_capA.append([facecount*3, facecount*3+1, facecount*3+2])
        facecount = facecount + 1
    vv_capA, _ = util.append_faces(v_capA, f_capA)

    #create cap-mesh-b

    v_capB, f_capB = [], []
    facecount = 0
    for capTriangle in cubicCapsB:
        for i in range(0, 3):
            v_capB.append([capTriangle[i][0], capTriangle[i][1], capTriangle[i][2]])
        f_capB.append([facecount*3, facecount*3+1, facecount*3+2])
        facecount = facecount + 1
    vv_capB, _ = util.append_faces(v_capB, f_capB)


    verticesA = openedMeshA.vertices
    facesA = openedMeshA.faces

    verticesB = openedMeshB.vertices
    facesB = openedMeshB.faces

    #eventually merge them with the opened mesh
    verticesA, facesA = util.append_faces([verticesA, v_a, vv_capA], [facesA, f_a, f_capA])
    verticesB, facesB = util.append_faces([verticesB, v_b, vv_capB], [facesB, f_b, f_capB])


    #verticesB, facesB = vv_capB, f_capB
    #vertices, faces = vv_cap, f_cap


    #processing will take some time
    sliced_mesh_A = Trimesh(vertices=verticesA, faces=facesA)
    verticesA, facesA = sliced_mesh_A.vertices.copy(), sliced_mesh_A.faces.copy()

    sliced_mesh_B = Trimesh(vertices=verticesB, faces=facesB)
    verticesB, facesB = sliced_mesh_B.vertices.copy(), sliced_mesh_B.faces.copy()

    return (Trimesh(vertices=verticesA, faces=facesA, process=False), Trimesh(vertices=verticesB, faces=facesB, process=False))
